"""Unit tests for sagemaker-automljob APIs."""

import datetime

import boto3

from moto import mock_aws

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html

ACCOUNT_ID = "123456789012"


@mock_aws
def test_create_auto_ml_job_v2():
    client = boto3.client("sagemaker", region_name="us-east-1")
    arn = client.create_auto_ml_job_v2(
        AutoMLJobName="testautomljob",
        AutoMLJobInputDataConfig=[
            {
                "ChannelType": "training",
                "ContentType": "ContentType",
                "CompressionType": "None",
                "DataSource": {
                    "S3DataSource": {"S3DataType": "S3Prefix", "S3Uri": "s3://data"}
                },
            },
        ],
        OutputDataConfig={"KmsKeyId": "kms", "S3OutputPath": "s3://output"},
        AutoMLProblemTypeConfig={
            "TextClassificationJobConfig": {
                "CompletionCriteria": {
                    "MaxCandidates": 123,
                    "MaxRuntimePerTrainingJobInSeconds": 123,
                    "MaxAutoMLJobRuntimeInSeconds": 123,
                },
                "ContentColumn": "content",
                "TargetLabelColumn": "target",
            },
        },
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole",
        Tags=[
            {"Key": "testkey", "Value": "testvalue"},
        ],
        SecurityConfig={
            "VolumeKmsKeyId": "testkmskeyid",
            "EnableInterContainerTrafficEncryption": True,
            "VpcConfig": {
                "SecurityGroupIds": [
                    "sg-12345678901234567",
                ],
                "Subnets": [
                    "subnet-12345678901234567",
                ],
            },
        },
        AutoMLJobObjective={"MetricName": "Accuracy"},
        ModelDeployConfig={
            "AutoGenerateEndpointName": False,
            "EndpointName": "testendpointname",
        },
        DataSplitConfig={"ValidationFraction": 0.2},
    )
    assert (
        arn["AutoMLJobArn"]
        == "arn:aws:sagemaker:us-east-1:123456789012:automl-job/testautomljob"
    )


@mock_aws
def test_describe_auto_ml_job_v2():
    client = boto3.client("sagemaker", region_name="us-west-2")
    client.create_auto_ml_job_v2(
        AutoMLJobName="testautomljob",
        AutoMLJobInputDataConfig=[
            {
                "ChannelType": "training",
                "ContentType": "ContentType",
                "CompressionType": "None",
                "DataSource": {
                    "S3DataSource": {"S3DataType": "S3Prefix", "S3Uri": "s3://data"}
                },
            },
        ],
        OutputDataConfig={"KmsKeyId": "kms", "S3OutputPath": "s3://output"},
        AutoMLProblemTypeConfig={
            "TextClassificationJobConfig": {
                "CompletionCriteria": {
                    "MaxCandidates": 123,
                    "MaxRuntimePerTrainingJobInSeconds": 123,
                    "MaxAutoMLJobRuntimeInSeconds": 123,
                },
                "ContentColumn": "content",
                "TargetLabelColumn": "target",
            },
        },
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole",
        Tags=[
            {"Key": "testkey", "Value": "testvalue"},
        ],
        SecurityConfig={
            "VolumeKmsKeyId": "testkmskeyid",
            "EnableInterContainerTrafficEncryption": True,
            "VpcConfig": {
                "SecurityGroupIds": [
                    "sg-12345678901234567",
                ],
                "Subnets": [
                    "subnet-12345678901234567",
                ],
            },
        },
        AutoMLJobObjective={"MetricName": "Accuracy"},
        ModelDeployConfig={
            "AutoGenerateEndpointName": False,
            "EndpointName": "testendpointname",
        },
        DataSplitConfig={"ValidationFraction": 0.3},
    )
    desc = client.describe_auto_ml_job_v2(AutoMLJobName="testautomljob")
    assert desc["AutoMLJobName"] == "testautomljob"
    assert (
        desc["AutoMLJobArn"]
        == "arn:aws:sagemaker:us-west-2:123456789012:automl-job/testautomljob"
    )
    assert desc["AutoMLJobInputDataConfig"] == [
        {
            "ChannelType": "training",
            "ContentType": "ContentType",
            "CompressionType": "None",
            "DataSource": {
                "S3DataSource": {"S3DataType": "S3Prefix", "S3Uri": "s3://data"}
            },
        },
    ]
    assert desc["OutputDataConfig"] == {
        "KmsKeyId": "kms",
        "S3OutputPath": "s3://output",
    }
    assert desc["RoleArn"] == f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole"
    assert desc["AutoMLJobObjective"] == {"MetricName": "Accuracy"}
    assert desc["AutoMLProblemTypeConfig"] == {
        "TextClassificationJobConfig": {
            "CompletionCriteria": {
                "MaxCandidates": 123,
                "MaxRuntimePerTrainingJobInSeconds": 123,
                "MaxAutoMLJobRuntimeInSeconds": 123,
            },
            "ContentColumn": "content",
            "TargetLabelColumn": "target",
        }
    }
    assert desc["AutoMLProblemTypeConfigName"] == "TextClassification"
    assert isinstance(desc["CreationTime"], datetime.datetime)
    assert isinstance(desc["EndTime"], datetime.datetime)
    assert isinstance(desc["LastModifiedTime"], datetime.datetime)
    assert desc["FailureReason"] == ""
    assert desc["PartialFailureReasons"] == [{"PartialFailureMessage": ""}]
    assert desc["BestCandidate"] == {
        "CandidateName": "best_candidate",
        "FinalAutoMLJobObjectiveMetric": {
            "Type": "Maximize",
            "MetricName": "Accuracy",
            "Value": 123,
            "StandardMetricName": "Accuracy",
        },
        "ObjectiveStatus": "Succeeded",
        "CandidateSteps": [
            {
                "CandidateStepType": "AWS::SageMaker::TrainingJob",
                "CandidateStepArn": "arn:aws:sagemaker:us-west-2:123456789012:training-job/candidate_step_name",
                "CandidateStepName": "candidate_step_name",
            },
        ],
        "CandidateStatus": "Completed",
        "InferenceContainers": [
            {
                "Image": "string",
                "ModelDataUrl": "string",
                "Environment": {"string": "string"},
            },
        ],
        "CreationTime": datetime.datetime(2024, 1, 1),
        "EndTime": datetime.datetime(2024, 1, 1),
        "LastModifiedTime": datetime.datetime(2024, 1, 1),
        "FailureReason": "string",
        "CandidateProperties": {
            "CandidateArtifactLocations": {
                "Explainability": "string",
                "ModelInsights": "string",
                "BacktestResults": "string",
            },
            "CandidateMetrics": [
                {
                    "MetricName": "Accuracy",
                    "Value": 123,
                    "Set": "Train",
                    "StandardMetricName": "Accuracy",
                },
            ],
        },
        "InferenceContainerDefinitions": {
            "string": [
                {
                    "Image": "string",
                    "ModelDataUrl": "string",
                    "Environment": {"string": "string"},
                },
            ]
        },
    }
    assert desc["AutoMLJobStatus"] == "InProgress"
    assert desc["AutoMLJobSecondaryStatus"] == "Completed"
    assert desc["AutoMLJobArtifacts"] == {
        "CandidateDefinitionNotebookLocation": "candidate/notebook/location",
        "DataExplorationNotebookLocation": "data/notebook/location",
    }
    assert desc["ResolvedAttributes"] == {
        "AutoMLJobObjective": {
            "MetricName": "Accuracy",
        },
        "CompletionCriteria": {
            "MaxCandidates": 123,
            "MaxRuntimePerTrainingJobInSeconds": 123,
            "MaxAutoMLJobRuntimeInSeconds": 123,
        },
        "AutoMLProblemTypeResolvedAttributes": {
            "SDK_UNKNOWN_MEMBER": {"name": "SDK_UNKNOWN_MEMBER"}
        },
    }
    assert desc["ModelDeployConfig"] == {
        "AutoGenerateEndpointName": False,
        "EndpointName": "testendpointname",
    }
    assert desc["DataSplitConfig"] == {"ValidationFraction": 0.3}
    assert desc["SecurityConfig"] == {
        "VolumeKmsKeyId": "testkmskeyid",
        "EnableInterContainerTrafficEncryption": True,
        "VpcConfig": {
            "SecurityGroupIds": [
                "sg-12345678901234567",
            ],
            "Subnets": [
                "subnet-12345678901234567",
            ],
        },
    }


@mock_aws
def test_describe_auto_ml_job_v2_defaults():
    client = boto3.client("sagemaker", region_name="us-west-2")
    client.create_auto_ml_job_v2(
        AutoMLJobName="testautomljob",
        AutoMLJobInputDataConfig=[
            {
                "ChannelType": "training",
                "ContentType": "ContentType",
                "CompressionType": "None",
                "DataSource": {
                    "S3DataSource": {"S3DataType": "S3Prefix", "S3Uri": "s3://data"}
                },
            },
        ],
        OutputDataConfig={"KmsKeyId": "kms", "S3OutputPath": "s3://output"},
        AutoMLProblemTypeConfig={
            "TextClassificationJobConfig": {
                "CompletionCriteria": {
                    "MaxCandidates": 123,
                    "MaxRuntimePerTrainingJobInSeconds": 123,
                    "MaxAutoMLJobRuntimeInSeconds": 123,
                },
                "ContentColumn": "content",
                "TargetLabelColumn": "target",
            },
        },
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole",
    )
    desc = client.describe_auto_ml_job_v2(AutoMLJobName="testautomljob")
    assert desc["AutoMLJobName"] == "testautomljob"
    assert (
        desc["AutoMLJobArn"]
        == "arn:aws:sagemaker:us-west-2:123456789012:automl-job/testautomljob"
    )
    assert desc["AutoMLJobInputDataConfig"] == [
        {
            "ChannelType": "training",
            "ContentType": "ContentType",
            "CompressionType": "None",
            "DataSource": {
                "S3DataSource": {"S3DataType": "S3Prefix", "S3Uri": "s3://data"}
            },
        },
    ]
    assert desc["OutputDataConfig"] == {
        "KmsKeyId": "kms",
        "S3OutputPath": "s3://output",
    }
    assert desc["RoleArn"] == f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole"
    assert desc["AutoMLJobObjective"] == {"MetricName": "Accuracy"}
    assert desc["AutoMLProblemTypeConfig"] == {
        "TextClassificationJobConfig": {
            "CompletionCriteria": {
                "MaxCandidates": 123,
                "MaxRuntimePerTrainingJobInSeconds": 123,
                "MaxAutoMLJobRuntimeInSeconds": 123,
            },
            "ContentColumn": "content",
            "TargetLabelColumn": "target",
        }
    }
    assert desc["AutoMLProblemTypeConfigName"] == "TextClassification"
    assert isinstance(desc["CreationTime"], datetime.datetime)
    assert isinstance(desc["EndTime"], datetime.datetime)
    assert isinstance(desc["LastModifiedTime"], datetime.datetime)
    assert desc["FailureReason"] == ""
    assert desc["PartialFailureReasons"] == [{"PartialFailureMessage": ""}]
    assert desc["BestCandidate"] == {
        "CandidateName": "best_candidate",
        "FinalAutoMLJobObjectiveMetric": {
            "Type": "Maximize",
            "MetricName": "Accuracy",
            "Value": 123,
            "StandardMetricName": "Accuracy",
        },
        "ObjectiveStatus": "Succeeded",
        "CandidateSteps": [
            {
                "CandidateStepType": "AWS::SageMaker::TrainingJob",
                "CandidateStepArn": "arn:aws:sagemaker:us-west-2:123456789012:training-job/candidate_step_name",
                "CandidateStepName": "candidate_step_name",
            },
        ],
        "CandidateStatus": "Completed",
        "InferenceContainers": [
            {
                "Image": "string",
                "ModelDataUrl": "string",
                "Environment": {"string": "string"},
            },
        ],
        "CreationTime": datetime.datetime(2024, 1, 1),
        "EndTime": datetime.datetime(2024, 1, 1),
        "LastModifiedTime": datetime.datetime(2024, 1, 1),
        "FailureReason": "string",
        "CandidateProperties": {
            "CandidateArtifactLocations": {
                "Explainability": "string",
                "ModelInsights": "string",
                "BacktestResults": "string",
            },
            "CandidateMetrics": [
                {
                    "MetricName": "Accuracy",
                    "Value": 123,
                    "Set": "Train",
                    "StandardMetricName": "Accuracy",
                },
            ],
        },
        "InferenceContainerDefinitions": {
            "string": [
                {
                    "Image": "string",
                    "ModelDataUrl": "string",
                    "Environment": {"string": "string"},
                },
            ]
        },
    }
    assert desc["AutoMLJobStatus"] == "InProgress"
    assert desc["AutoMLJobSecondaryStatus"] == "Completed"
    assert desc["AutoMLJobArtifacts"] == {
        "CandidateDefinitionNotebookLocation": "candidate/notebook/location",
        "DataExplorationNotebookLocation": "data/notebook/location",
    }
    assert desc["ResolvedAttributes"] == {
        "AutoMLJobObjective": {
            "MetricName": "Accuracy",
        },
        "CompletionCriteria": {
            "MaxCandidates": 123,
            "MaxRuntimePerTrainingJobInSeconds": 123,
            "MaxAutoMLJobRuntimeInSeconds": 123,
        },
        "AutoMLProblemTypeResolvedAttributes": {
            "SDK_UNKNOWN_MEMBER": {"name": "SDK_UNKNOWN_MEMBER"}
        },
    }
    assert desc["ModelDeployConfig"] == {
        "AutoGenerateEndpointName": False,
        "EndpointName": "EndpointName",
    }
    assert desc["DataSplitConfig"] == {"ValidationFraction": 0.2}


@mock_aws
def test_list_auto_ml_jobs():
    client = boto3.client("sagemaker", region_name="ap-southeast-1")
    client.create_auto_ml_job_v2(
        AutoMLJobName="testautomljob",
        AutoMLJobInputDataConfig=[
            {
                "ChannelType": "training",
                "ContentType": "ContentType",
                "CompressionType": "None",
                "DataSource": {
                    "S3DataSource": {"S3DataType": "S3Prefix", "S3Uri": "s3://data"}
                },
            },
        ],
        OutputDataConfig={"KmsKeyId": "kms", "S3OutputPath": "s3://output"},
        AutoMLProblemTypeConfig={
            "TextClassificationJobConfig": {
                "CompletionCriteria": {
                    "MaxCandidates": 123,
                    "MaxRuntimePerTrainingJobInSeconds": 123,
                    "MaxAutoMLJobRuntimeInSeconds": 123,
                },
                "ContentColumn": "content",
                "TargetLabelColumn": "target",
            },
        },
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole",
    )
    client.create_auto_ml_job_v2(
        AutoMLJobName="testautomljob2",
        AutoMLJobInputDataConfig=[
            {
                "ChannelType": "training",
                "ContentType": "ContentType",
                "CompressionType": "None",
                "DataSource": {
                    "S3DataSource": {"S3DataType": "S3Prefix", "S3Uri": "s3://data"}
                },
            },
        ],
        OutputDataConfig={"KmsKeyId": "kms", "S3OutputPath": "s3://output"},
        AutoMLProblemTypeConfig={
            "TabularJobConfig": {
                "CandidateGenerationConfig": {
                    "AlgorithmsConfig": [
                        {"AutoMLAlgorithms": ["xgboost"]},
                    ]
                },
                "CompletionCriteria": {
                    "MaxCandidates": 123,
                    "MaxRuntimePerTrainingJobInSeconds": 123,
                    "MaxAutoMLJobRuntimeInSeconds": 123,
                },
                "FeatureSpecificationS3Uri": "string",
                "Mode": "AUTO",
                "GenerateCandidateDefinitionsOnly": True,
                "ProblemType": "BinaryClassification",
                "TargetAttributeName": "string",
                "SampleWeightAttributeName": "string",
            }
        },
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole",
    )
    resp = client.list_auto_ml_jobs()
    assert len(resp["AutoMLJobSummaries"]) == 2
    assert resp["AutoMLJobSummaries"][0]["AutoMLJobName"] == "testautomljob"
    assert (
        resp["AutoMLJobSummaries"][0]["AutoMLJobArn"]
        == f"arn:aws:sagemaker:ap-southeast-1:{ACCOUNT_ID}:automl-job/testautomljob"
    )
    assert resp["AutoMLJobSummaries"][0]["AutoMLJobStatus"] == "InProgress"
    assert resp["AutoMLJobSummaries"][0]["AutoMLJobSecondaryStatus"] == "Completed"
    assert isinstance(resp["AutoMLJobSummaries"][0]["CreationTime"], datetime.datetime)
    assert isinstance(resp["AutoMLJobSummaries"][0]["EndTime"], datetime.datetime)
    assert resp["AutoMLJobSummaries"][0]["FailureReason"] == ""
    assert resp["AutoMLJobSummaries"][0]["PartialFailureReasons"] == [
        {"PartialFailureMessage": ""}
    ]
    assert resp["AutoMLJobSummaries"][1]["AutoMLJobName"] == "testautomljob2"
    assert (
        resp["AutoMLJobSummaries"][1]["AutoMLJobArn"]
        == f"arn:aws:sagemaker:ap-southeast-1:{ACCOUNT_ID}:automl-job/testautomljob2"
    )
    assert resp["AutoMLJobSummaries"][1]["AutoMLJobStatus"] == "InProgress"
    assert resp["AutoMLJobSummaries"][1]["AutoMLJobSecondaryStatus"] == "Completed"
    assert isinstance(resp["AutoMLJobSummaries"][1]["CreationTime"], datetime.datetime)
    assert isinstance(resp["AutoMLJobSummaries"][1]["EndTime"], datetime.datetime)
    assert resp["AutoMLJobSummaries"][1]["FailureReason"] == ""
    assert resp["AutoMLJobSummaries"][1]["PartialFailureReasons"] == [
        {"PartialFailureMessage": ""}
    ]


@mock_aws
def test_list_auto_ml_jobs_filters():
    client = boto3.client("sagemaker", region_name="ap-southeast-1")
    client.create_auto_ml_job_v2(
        AutoMLJobName="testautomljob",
        AutoMLJobInputDataConfig=[
            {
                "ChannelType": "training",
                "ContentType": "ContentType",
                "CompressionType": "None",
                "DataSource": {
                    "S3DataSource": {"S3DataType": "S3Prefix", "S3Uri": "s3://data"}
                },
            },
        ],
        OutputDataConfig={"KmsKeyId": "kms", "S3OutputPath": "s3://output"},
        AutoMLProblemTypeConfig={
            "ImageClassificationJobConfig": {
                "CompletionCriteria": {
                    "MaxCandidates": 123,
                    "MaxRuntimePerTrainingJobInSeconds": 123,
                    "MaxAutoMLJobRuntimeInSeconds": 123,
                }
            },
        },
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole",
    )

    client.create_auto_ml_job_v2(
        AutoMLJobName="testautomljob2",
        AutoMLJobInputDataConfig=[
            {
                "ChannelType": "training",
                "ContentType": "ContentType",
                "CompressionType": "None",
                "DataSource": {
                    "S3DataSource": {"S3DataType": "S3Prefix", "S3Uri": "s3://data"}
                },
            },
        ],
        OutputDataConfig={"KmsKeyId": "kms", "S3OutputPath": "s3://output"},
        AutoMLProblemTypeConfig={
            "TimeSeriesForecastingJobConfig": {
                "FeatureSpecificationS3Uri": "string",
                "CompletionCriteria": {
                    "MaxCandidates": 123,
                    "MaxRuntimePerTrainingJobInSeconds": 123,
                    "MaxAutoMLJobRuntimeInSeconds": 123,
                },
                "ForecastFrequency": "string",
                "ForecastHorizon": 123,
                "ForecastQuantiles": [
                    "string",
                ],
                "Transformations": {
                    "Filling": {"string": {"string": "string"}},
                    "Aggregation": {"string": "sum"},
                },
                "TimeSeriesConfig": {
                    "TargetAttributeName": "string",
                    "TimestampAttributeName": "string",
                    "ItemIdentifierAttributeName": "string",
                    "GroupingAttributeNames": [
                        "string",
                    ],
                },
                "HolidayConfig": [
                    {"CountryCode": "string"},
                ],
                "CandidateGenerationConfig": {
                    "AlgorithmsConfig": [
                        {"AutoMLAlgorithms": ["xgboost"]},
                    ]
                },
            },
        },
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole",
    )

    client.create_auto_ml_job_v2(
        AutoMLJobName="test3automljob",
        AutoMLJobInputDataConfig=[
            {
                "ChannelType": "training",
                "ContentType": "ContentType",
                "CompressionType": "None",
                "DataSource": {
                    "S3DataSource": {"S3DataType": "S3Prefix", "S3Uri": "s3://data"}
                },
            },
        ],
        OutputDataConfig={"KmsKeyId": "kms", "S3OutputPath": "s3://output"},
        AutoMLProblemTypeConfig={
            "TextGenerationJobConfig": {
                "CompletionCriteria": {
                    "MaxCandidates": 123,
                    "MaxRuntimePerTrainingJobInSeconds": 123,
                    "MaxAutoMLJobRuntimeInSeconds": 123,
                },
                "BaseModelName": "string",
                "TextGenerationHyperParameters": {"string": "string"},
                "ModelAccessConfig": {"AcceptEula": True},
            }
        },
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole",
    )

    resp = client.list_auto_ml_jobs(
        CreationTimeAfter=datetime.datetime(2024, 1, 1),
        CreationTimeBefore=datetime.datetime(2099, 1, 1),
        LastModifiedTimeAfter=datetime.datetime(2024, 1, 1),
        LastModifiedTimeBefore=datetime.datetime(2099, 1, 1),
        NameContains="testautomljob",
        StatusEquals="InProgress",
        SortOrder="Ascending",
        SortBy="Name",
    )
    assert len(resp["AutoMLJobSummaries"]) == 2
    assert resp["AutoMLJobSummaries"][0]["AutoMLJobName"] == "testautomljob"
    assert (
        resp["AutoMLJobSummaries"][0]["AutoMLJobArn"]
        == f"arn:aws:sagemaker:ap-southeast-1:{ACCOUNT_ID}:automl-job/testautomljob"
    )
    assert resp["AutoMLJobSummaries"][0]["AutoMLJobStatus"] == "InProgress"
    assert resp["AutoMLJobSummaries"][0]["AutoMLJobSecondaryStatus"] == "Completed"
    assert isinstance(resp["AutoMLJobSummaries"][0]["CreationTime"], datetime.datetime)
    assert isinstance(resp["AutoMLJobSummaries"][0]["EndTime"], datetime.datetime)
    assert resp["AutoMLJobSummaries"][0]["FailureReason"] == ""
    assert resp["AutoMLJobSummaries"][0]["PartialFailureReasons"] == [
        {"PartialFailureMessage": ""}
    ]
    assert resp["AutoMLJobSummaries"][1]["AutoMLJobName"] == "testautomljob2"
    assert (
        resp["AutoMLJobSummaries"][1]["AutoMLJobArn"]
        == f"arn:aws:sagemaker:ap-southeast-1:{ACCOUNT_ID}:automl-job/testautomljob2"
    )
    assert resp["AutoMLJobSummaries"][1]["AutoMLJobStatus"] == "InProgress"
    assert resp["AutoMLJobSummaries"][1]["AutoMLJobSecondaryStatus"] == "Completed"
    assert isinstance(resp["AutoMLJobSummaries"][1]["CreationTime"], datetime.datetime)
    assert isinstance(resp["AutoMLJobSummaries"][1]["EndTime"], datetime.datetime)
    assert resp["AutoMLJobSummaries"][1]["FailureReason"] == ""
    assert resp["AutoMLJobSummaries"][1]["PartialFailureReasons"] == [
        {"PartialFailureMessage": ""}
    ]


@mock_aws
def test_stop_auto_ml_job():
    client = boto3.client("sagemaker", region_name="ap-southeast-1")
    client.create_auto_ml_job_v2(
        AutoMLJobName="testautomljob",
        AutoMLJobInputDataConfig=[
            {
                "ChannelType": "training",
                "ContentType": "ContentType",
                "CompressionType": "None",
                "DataSource": {
                    "S3DataSource": {"S3DataType": "S3Prefix", "S3Uri": "s3://data"}
                },
            },
        ],
        OutputDataConfig={"KmsKeyId": "kms", "S3OutputPath": "s3://output"},
        AutoMLProblemTypeConfig={
            "ImageClassificationJobConfig": {
                "CompletionCriteria": {
                    "MaxCandidates": 123,
                    "MaxRuntimePerTrainingJobInSeconds": 123,
                    "MaxAutoMLJobRuntimeInSeconds": 123,
                }
            },
        },
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole",
    )
    initial_resp = client.describe_auto_ml_job_v2(AutoMLJobName="testautomljob")
    assert initial_resp["AutoMLJobStatus"] == "InProgress"
    client.stop_auto_ml_job(AutoMLJobName="testautomljob")
    resp = client.describe_auto_ml_job_v2(AutoMLJobName="testautomljob")
    assert resp["AutoMLJobStatus"] == "Stopped"


@mock_aws
def test_tag_auto_ml_job():
    client = boto3.client("sagemaker", region_name="ap-southeast-1")
    arn = client.create_auto_ml_job_v2(
        AutoMLJobName="testautomljob",
        AutoMLJobInputDataConfig=[
            {
                "ChannelType": "training",
                "ContentType": "ContentType",
                "CompressionType": "None",
                "DataSource": {
                    "S3DataSource": {"S3DataType": "S3Prefix", "S3Uri": "s3://data"}
                },
            },
        ],
        OutputDataConfig={"KmsKeyId": "kms", "S3OutputPath": "s3://output"},
        AutoMLProblemTypeConfig={
            "ImageClassificationJobConfig": {
                "CompletionCriteria": {
                    "MaxCandidates": 123,
                    "MaxRuntimePerTrainingJobInSeconds": 123,
                    "MaxAutoMLJobRuntimeInSeconds": 123,
                }
            },
        },
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole",
        Tags=[{"Key": "testkey", "Value": "testvalue"}],
    )
    initial_resp = client.list_tags(ResourceArn=arn["AutoMLJobArn"])
    assert initial_resp["Tags"] == [{"Key": "testkey", "Value": "testvalue"}]
    client.add_tags(
        ResourceArn=arn["AutoMLJobArn"],
        Tags=[{"Key": "testkey2", "Value": "testvalue2"}],
    )
    resp = client.list_tags(ResourceArn=arn["AutoMLJobArn"])
    assert resp["Tags"] == [
        {"Key": "testkey", "Value": "testvalue"},
        {"Key": "testkey2", "Value": "testvalue2"},
    ]
    client.delete_tags(ResourceArn=arn["AutoMLJobArn"], TagKeys=["testkey"]) == [
        {"Key": "testkey2", "Value": "testvalue2"}
    ]



from typing import Any, Dict


def clean_dict(obj: Dict[str, Any]) -> Dict[str, Any]:  # type: ignore[misc]
    return {
        key: value for key, value in obj.items() if value is not None and value != []
    }



"""Unit tests for pinpoint-supported APIs."""

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@mock_aws
def test_create_app():
    client = boto3.client("pinpoint", region_name="us-east-1")
    resp = client.create_app(CreateApplicationRequest={"Name": "myfirstapp"})

    assert "ApplicationResponse" in resp
    assert "Arn" in resp["ApplicationResponse"]
    assert "us-east-1" in resp["ApplicationResponse"]["Arn"]
    assert "Id" in resp["ApplicationResponse"]
    assert resp["ApplicationResponse"]["Name"] == "myfirstapp"
    assert "CreationDate" in resp["ApplicationResponse"]


@mock_aws
def test_delete_app():
    client = boto3.client("pinpoint", region_name="ap-southeast-1")
    creation = client.create_app(CreateApplicationRequest={"Name": "myfirstapp"})[
        "ApplicationResponse"
    ]
    app_id = creation["Id"]

    deletion = client.delete_app(ApplicationId=app_id)["ApplicationResponse"]
    assert deletion == creation

    with pytest.raises(ClientError) as exc:
        client.get_app(ApplicationId=app_id)
    err = exc.value.response["Error"]
    assert err["Code"] == "NotFoundException"
    assert err["Message"] == "Application not found"


@mock_aws
def test_get_app():
    client = boto3.client("pinpoint", region_name="eu-west-1")
    resp = client.create_app(CreateApplicationRequest={"Name": "myfirstapp"})
    app_id = resp["ApplicationResponse"]["Id"]

    resp = client.get_app(ApplicationId=app_id)

    assert "ApplicationResponse" in resp
    assert "Arn" in resp["ApplicationResponse"]
    assert "eu-west-1" in resp["ApplicationResponse"]["Arn"]
    assert "Id" in resp["ApplicationResponse"]
    assert resp["ApplicationResponse"]["Name"] == "myfirstapp"
    assert "CreationDate" in resp["ApplicationResponse"]


@mock_aws
def test_get_apps_initial():
    client = boto3.client("pinpoint", region_name="us-east-1")
    resp = client.get_apps()

    assert "ApplicationsResponse" in resp
    assert resp["ApplicationsResponse"] == {"Item": []}


@mock_aws
def test_get_apps():
    client = boto3.client("pinpoint", region_name="us-east-1")
    resp = client.create_app(CreateApplicationRequest={"Name": "myfirstapp"})
    app_id = resp["ApplicationResponse"]["Id"]

    resp = client.get_apps()

    assert len(resp["ApplicationsResponse"]["Item"]) == 1
    assert "Arn" in resp["ApplicationsResponse"]["Item"][0]
    assert resp["ApplicationsResponse"]["Item"][0]["Id"] == app_id
    assert resp["ApplicationsResponse"]["Item"][0]["Name"] == "myfirstapp"
    assert "CreationDate" in resp["ApplicationsResponse"]["Item"][0]


@mock_aws
def test_update_application_settings():
    client = boto3.client("pinpoint", region_name="eu-west-1")
    resp = client.create_app(CreateApplicationRequest={"Name": "myfirstapp"})
    app_id = resp["ApplicationResponse"]["Id"]

    resp = client.update_application_settings(
        ApplicationId=app_id,
        WriteApplicationSettingsRequest={
            "CampaignHook": {"LambdaFunctionName": "lfn"},
            "CloudWatchMetricsEnabled": True,
            "EventTaggingEnabled": True,
            "Limits": {"Daily": 42},
        },
    )

    assert "ApplicationSettingsResource" in resp
    app_settings = resp["ApplicationSettingsResource"]
    assert app_settings["ApplicationId"] == app_id
    assert app_settings["CampaignHook"] == {"LambdaFunctionName": "lfn"}
    assert app_settings["Limits"] == {"Daily": 42}
    assert "LastModifiedDate" in app_settings


@mock_aws
def test_get_application_settings():
    client = boto3.client("pinpoint", region_name="ap-southeast-1")
    resp = client.create_app(CreateApplicationRequest={"Name": "myfirstapp"})
    app_id = resp["ApplicationResponse"]["Id"]

    client.update_application_settings(
        ApplicationId=app_id,
        WriteApplicationSettingsRequest={
            "CampaignHook": {"LambdaFunctionName": "lfn"},
            "CloudWatchMetricsEnabled": True,
            "EventTaggingEnabled": True,
            "Limits": {"Daily": 42},
        },
    )

    resp = client.get_application_settings(ApplicationId=app_id)

    assert "ApplicationSettingsResource" in resp
    app_settings = resp["ApplicationSettingsResource"]
    assert app_settings["ApplicationId"] == app_id
    assert app_settings["CampaignHook"] == {"LambdaFunctionName": "lfn"}
    assert app_settings["Limits"] == {"Daily": 42}
    assert "LastModifiedDate" in app_settings



from functools import partial

from moto.utilities.utils import PARTITION_NAMES, get_partition


def make_arn(
    name: str, account_id: str, region_name: str, _id: str, scope: str, resource: str
) -> str:
    if scope == "REGIONAL":
        scope = "regional"
    elif scope == "CLOUDFRONT":
        scope = "global"

    if region_name in PARTITION_NAMES:
        region_name = "global"
    partition = (
        region_name if region_name in PARTITION_NAMES else get_partition(region_name)
    )

    return f"arn:{partition}:wafv2:{region_name}:{account_id}:{scope}/{resource}/{name}/{_id}"


make_arn_for_wacl = partial(make_arn, resource="webacl")
make_arn_for_ip_set = partial(make_arn, resource="ipset")
make_arn_for_logging_configuration = partial(make_arn, resource="loggingconfiguration")
make_arn_for_rule_group = partial(make_arn, resource="rulegroup")



"""Unit tests for rekognition-supported APIs."""

import random
import string

import boto3

from moto import mock_aws
from moto.moto_api._internal import mock_random

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@mock_aws
def test_start_face_search():
    client = boto3.client("rekognition", region_name="ap-southeast-1")
    collection_id = "collection_id"
    video = {
        "S3Object": {
            "Bucket": "bucket",
            "Name": "key",
        }
    }

    resp = client.start_face_search(CollectionId=collection_id, Video=video)

    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200
    assert "JobId" in resp


@mock_aws
def test_start_text_detection():
    client = boto3.client("rekognition", region_name="ap-southeast-1")
    video = {
        "S3Object": {
            "Bucket": "bucket",
            "Name": "key",
        }
    }

    resp = client.start_text_detection(Video=video)

    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200
    assert "JobId" in resp


@mock_aws
def test_compare_faces():
    client = boto3.client("rekognition", region_name="ap-southeast-1")
    sourceimage = {
        "S3Object": {"Bucket": "string", "Name": "string", "Version": "string"}
    }
    targetimage = {
        "S3Object": {"Bucket": "string", "Name": "string", "Version": "string"}
    }

    resp = client.compare_faces(
        SimilarityThreshold=80, SourceImage=sourceimage, TargetImage=targetimage
    )

    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200
    assert "FaceMatches" in resp


@mock_aws
def test_detect_labels():
    client = boto3.client("rekognition", region_name="ap-southeast-1")

    resp = client.detect_labels(
        Image={"S3Object": {"Bucket": "string", "Name": "name.jpg"}}, MaxLabels=10
    )

    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200
    assert "Labels" in resp


@mock_aws
def test_detect_text():
    client = boto3.client("rekognition", region_name="ap-southeast-1")

    resp = client.detect_text(
        Image={"S3Object": {"Bucket": "string", "Name": "name.jpg"}}
    )

    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200
    assert "TextDetections" in resp


@mock_aws
def test_get_face_search():
    client = boto3.client("rekognition", region_name="us-east-2")
    job_id = "".join(
        random.choice(string.ascii_uppercase + string.digits) for _ in range(64)
    )

    resp = client.get_face_search(JobId=job_id)

    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200

    assert resp["JobStatus"] == "SUCCEEDED"
    assert (
        resp["Persons"][0]["FaceMatches"][0]["Face"]["ExternalImageId"] == "Dave_Bloggs"
    )


@mock_aws
def test_get_text_detection():
    client = boto3.client("rekognition", region_name="us-east-2")
    job_id = "".join(
        random.choice(string.ascii_uppercase + string.digits) for _ in range(64)
    )

    resp = client.get_text_detection(JobId=job_id)

    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200

    assert resp["JobStatus"] == "SUCCEEDED"
    assert resp["TextDetections"][0]["TextDetection"]["DetectedText"] == "Hello world"


@mock_aws
def test_detect_custom_labels():
    client = boto3.client("rekognition", region_name="us-east-2")
    resp = client.detect_custom_labels(
        Image={"S3Object": {"Bucket": "string", "Name": "name.jpg"}},
        MaxResults=10,
        MinConfidence=80,
        ProjectVersionArn=mock_random.get_random_string(),
    )
    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200
    assert resp["CustomLabels"][0]["Name"] == "MyLogo"



import json

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws

invalid_policy_document_test_cases = [
    {
        "document": "This is not a json document",
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
            }
        },
        "error_message": "Policy document must be version 2012-10-17 or greater.",
    },
    {
        "document": {
            "Version": "2008-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
            },
        },
        "error_message": "Policy document must be version 2012-10-17 or greater.",
    },
    {
        "document": {
            "Version": "2013-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {"Version": "2012-10-17"},
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {"Version": "2012-10-17", "Statement": ["afd"]},
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
            },
            "Extra field": "value",
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
                "Extra field": "value",
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Id": ["cd3a324d2343d942772346-34234234423404-4c2242343242349d1642ee"],
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Id": {},
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "invalid",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "invalid",
                "Resource": "arn:aws:s3:::example_bucket",
            },
        },
        "error_message": "Actions/Conditions must be prefaced by a vendor, e.g., iam, sdb, ec2, etc.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "NotAction": "",
                "Resource": "arn:aws:s3:::example_bucket",
            },
        },
        "error_message": "Actions/Conditions must be prefaced by a vendor, e.g., iam, sdb, ec2, etc.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "a a:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
            },
        },
        "error_message": "Vendor a a is not valid",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:List:Bucket",
                "Resource": "arn:aws:s3:::example_bucket",
            },
        },
        "error_message": "Actions/Condition can contain only one colon.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": "s3s:ListBucket",
                    "Resource": "arn:aws:s3:::example_bucket",
                },
                {
                    "Effect": "Allow",
                    "Action": "s:3s:ListBucket",
                    "Resource": "arn:aws:s3:::example_bucket",
                },
            ],
        },
        "error_message": "Actions/Condition can contain only one colon.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "invalid resource",
            },
        },
        "error_message": 'Resource invalid resource must be in ARN format or "*".',
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "EnableDisableHongKong",
                    "Effect": "Allow",
                    "Action": ["account:EnableRegion", "account:DisableRegion"],
                    "Resource": "",
                    "Condition": {
                        "StringEquals": {"account:TargetRegion": "ap-east-1"}
                    },
                },
                {
                    "Sid": "ViewConsole",
                    "Effect": "Allow",
                    "Action": ["aws-portal:ViewAccount", "account:ListRegions"],
                    "Resource": "",
                },
            ],
        },
        "error_message": 'Resource  must be in ARN format or "*".',
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s:3:ListBucket",
                "Resource": "sdfsadf",
            },
        },
        "error_message": 'Resource sdfsadf must be in ARN format or "*".',
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": ["adf"],
            },
        },
        "error_message": 'Resource adf must be in ARN format or "*".',
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {"Effect": "Allow", "Action": "s3:ListBucket", "Resource": ""},
        },
        "error_message": 'Resource  must be in ARN format or "*".',
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "NotAction": "s3s:ListBucket",
                "Resource": "a:bsdfdsafsad",
            },
        },
        "error_message": 'Partition "bsdfdsafsad" is not valid for resource "arn:bsdfdsafsad:*:*:*:*".',
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "NotAction": "s3s:ListBucket",
                "Resource": "a:b:cadfsdf",
            },
        },
        "error_message": 'Partition "b" is not valid for resource "arn:b:cadfsdf:*:*:*".',
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "NotAction": "s3s:ListBucket",
                "Resource": "a:b:c:d:e:f:g:h",
            },
        },
        "error_message": 'Partition "b" is not valid for resource "arn:b:c:d:e:f:g:h".',
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "aws:s3:::example_bucket",
            },
        },
        "error_message": 'Partition "s3" is not valid for resource "arn:s3:::example_bucket:*".',
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": [
                    "arn:error:s3:::example_bucket",
                    "arn:error:s3::example_bucket",
                ],
            },
        },
        "error_message": 'Partition "error" is not valid for resource "arn:error:s3:::example_bucket".',
    },
    {
        "document": {"Version": "2012-10-17", "Statement": []},
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {"Effect": "Allow", "Action": "s3:ListBucket"},
        },
        "error_message": "Policy statement must contain resources.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {"Effect": "Allow", "Action": "s3:ListBucket", "Resource": []},
        },
        "error_message": "Policy statement must contain resources.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {"Effect": "Allow", "Action": "invalid"},
        },
        "error_message": "Actions/Conditions must be prefaced by a vendor, e.g., iam, sdb, ec2, etc.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {"Effect": "Allow", "Resource": "arn:aws:s3:::example_bucket"},
        },
        "error_message": "Policy statement must contain actions.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {"Version": "2012-10-17", "Statement": {"Effect": "Allow"}},
        "error_message": "Policy statement must contain actions.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": [],
                "Resource": "arn:aws:s3:::example_bucket",
            },
        },
        "error_message": "Policy statement must contain actions.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": [
                {"Effect": "Deny"},
                {
                    "Effect": "Allow",
                    "Action": "s3:ListBucket",
                    "Resource": "arn:aws:s3:::example_bucket",
                },
            ],
        },
        "error_message": "Policy statement must contain actions.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:iam:::example_bucket",
            },
        },
        "error_message": 'IAM resource path must either be "*" or start with user/, federated-user/, role/, group/, instance-profile/, mfa/, server-certificate/, policy/, sms-mfa/, saml-provider/, oidc-provider/, report/, access-report/.',
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3::example_bucket",
            },
        },
        "error_message": "The policy failed legacy parsing",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {"Effect": "Allow", "Resource": "arn:aws:s3::example_bucket"},
        },
        "error_message": "The policy failed legacy parsing",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws",
            },
        },
        "error_message": "Resource vendor must be fully qualified and cannot contain regexes.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": {"a": "arn:aws:s3:::example_bucket"},
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Deny",
                "Action": "s3:ListBucket",
                "Resource": ["adfdf", {}],
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "NotAction": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
                "NotResource": [],
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Deny",
                "Action": [[]],
                "Resource": "arn:aws:s3:::example_bucket",
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "NotAction": "s3s:ListBucket",
                "Action": [],
                "Resource": "arn:aws:s3:::example_bucket",
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": {},
                "Resource": "arn:aws:s3:::example_bucket",
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
                "Condition": [],
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
                "Condition": "a",
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
                "Condition": {"a": "b"},
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
                "Condition": {"DateGreaterThan": "b"},
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
                "Condition": {"DateGreaterThan": []},
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
                "Condition": {"DateGreaterThan": {"a": {}}},
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Resource": "arn:aws:s3:::example_bucket",
                "Condition": {"DateGreaterThan": {"a": {}}},
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
                "Condition": {"x": {"a": "1"}},
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
                "Condition": {"ForAnyValue::StringEqualsIfExists": {"a": "asf"}},
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
                "Condition": [
                    {"ForAllValues:StringEquals": {"aws:TagKeys": "Department"}}
                ],
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:iam:us-east-1::example_bucket",
            },
        },
        "error_message": "IAM resource arn:aws:iam:us-east-1::example_bucket cannot contain region information.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:us-east-1::example_bucket",
            },
        },
        "error_message": "Resource arn:aws:s3:us-east-1::example_bucket can not contain region information.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Sid": {},
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Sid": [],
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "sdf",
                    "Effect": "Allow",
                    "Action": "s3:ListBucket",
                    "Resource": "arn:aws:s3:::example_bucket",
                },
                {"Sid": "sdf", "Effect": "Allow"},
            ],
        },
        "error_message": "Statement IDs (SID) in a single policy must be unique.",
    },
    {
        "document": {
            "Statement": [
                {
                    "Sid": "sdf",
                    "Effect": "Allow",
                    "Action": "s3:ListBucket",
                    "Resource": "arn:aws:s3:::example_bucket",
                },
                {"Sid": "sdf", "Effect": "Allow"},
            ]
        },
        "error_message": "Policy document must be version 2012-10-17 or greater.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "NotAction": "s3:ListBucket",
                "Action": "iam:dsf",
                "Resource": "arn:aws:s3:::example_bucket",
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
                "NotResource": "*",
            },
        },
        "error_message": "Syntax errors in policy.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "denY",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
            },
        },
        "error_message": "The policy failed legacy parsing",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
                "Condition": {"DateGreaterThan": {"a": "sdfdsf"}},
            },
        },
        "error_message": "The policy failed legacy parsing",
    },
    {
        "document": {
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
                "Condition": {"DateGreaterThan": {"a": "sdfdsf"}},
            }
        },
        "error_message": "Policy document must be version 2012-10-17 or greater.",
    },
    {
        "document": {
            "Statement": {
                "Effect": "denY",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
            }
        },
        "error_message": "Policy document must be version 2012-10-17 or greater.",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Condition": {"DateGreaterThan": {"a": "sdfdsf"}},
            },
        },
        "error_message": "The policy failed legacy parsing",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "NotAction": "s3:ListBucket",
                "Resource": "arn:aws::::example_bucket",
            },
        },
        "error_message": "The policy failed legacy parsing",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "allow",
                "Resource": "arn:aws:s3:us-east-1::example_bucket",
            },
        },
        "error_message": "The policy failed legacy parsing",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "sdf",
                    "Effect": "aLLow",
                    "Action": "s3:ListBucket",
                    "Resource": "arn:aws:s3:::example_bucket",
                },
                {"Sid": "sdf", "Effect": "Allow"},
            ],
        },
        "error_message": "The policy failed legacy parsing",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "NotResource": "arn:aws:s3::example_bucket",
            },
        },
        "error_message": "The policy failed legacy parsing",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
                "Condition": {"DateLessThanEquals": {"a": "234-13"}},
            },
        },
        "error_message": "The policy failed legacy parsing",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
                "Condition": {
                    "DateLessThanEquals": {"a": "2016-12-13t2:00:00.593194+1"}
                },
            },
        },
        "error_message": "The policy failed legacy parsing",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
                "Condition": {
                    "DateLessThanEquals": {"a": "2016-12-13t2:00:00.1999999999+10:59"}
                },
            },
        },
        "error_message": "The policy failed legacy parsing",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
                "Condition": {"DateLessThan": {"a": "9223372036854775808"}},
            },
        },
        "error_message": "The policy failed legacy parsing",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:error:s3:::example_bucket",
                "Condition": {"DateGreaterThan": {"a": "sdfdsf"}},
            },
        },
        "error_message": "The policy failed legacy parsing",
    },
    {
        "document": {
            "Version": "2012-10-17",
            "Statement": {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws::fdsasf",
            },
        },
        "error_message": "The policy failed legacy parsing",
    },
]

valid_policy_documents = [
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": ["arn:aws:s3:::example_bucket"],
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "iam: asdf safdsf af ",
            "Resource": "arn:aws:s3:::example_bucket",
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": ["arn:aws:s3:::example_bucket", "*"],
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "*",
            "Resource": "arn:aws:s3:::example_bucket",
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
            }
        ],
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "service-prefix:action-name",
            "Resource": "*",
            "Condition": {
                "DateGreaterThan": {"aws:CurrentTime": "2017-07-01T00:00:00Z"},
                "DateLessThan": {"aws:CurrentTime": "2017-12-31T23:59:59Z"},
            },
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "fsx:ListBucket",
            "Resource": "arn:aws:s3:::example_bucket",
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": "arn:aws:iam:::user/example_bucket",
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": "arn:aws:s33:::example_bucket",
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": "arn:aws:fdsasf",
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": "arn:aws:s3:::example_bucket",
            "Condition": {},
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": "arn:aws:s3:::example_bucket",
            "Condition": {"ForAllValues:StringEquals": {"aws:TagKeys": "Department"}},
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": "arn:aws:cloudwatch:us-east-1::example_bucket",
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": "arn:aws:ec2:us-east-1::example_bucket",
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": "arn:aws:invalid-service:::example_bucket",
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": "arn:aws:invalid-service:us-east-1::example_bucket",
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": "arn:aws:s3:::example_bucket",
            "Condition": {
                "DateGreaterThan": {"aws:CurrentTime": "2017-07-01T00:00:00Z"},
                "DateLessThan": {"aws:CurrentTime": "2017-12-31T23:59:59Z"},
            },
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": "arn:aws:s3:::example_bucket",
            "Condition": {"DateGreaterThan": {}},
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": "arn:aws:s3:::example_bucket",
            "Condition": {"DateGreaterThan": {"a": []}},
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": "arn:aws:s3:::example_bucket",
            "Condition": {"a": {}},
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Sid": "dsfsdfsdfsdfsdfsadfsd",
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": "arn:aws:s3:::example_bucket",
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "ConsoleDisplay",
                "Effect": "Allow",
                "Action": [
                    "iam:GetRole",
                    "iam:GetUser",
                    "iam:ListRoles",
                    "iam:ListRoleTags",
                    "iam:ListUsers",
                    "iam:ListUserTags",
                ],
                "Resource": "*",
            },
            {
                "Sid": "AddTag",
                "Effect": "Allow",
                "Action": ["iam:TagUser", "iam:TagRole"],
                "Resource": "*",
                "Condition": {
                    "StringEquals": {"aws:RequestTag/CostCenter": ["A-123", "B-456"]},
                    "ForAllValues:StringEquals": {"aws:TagKeys": "CostCenter"},
                },
            },
        ],
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "NotAction": "s3:ListBucket",
            "Resource": "arn:aws:s3:::example_bucket",
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Deny",
            "Action": "s3:*",
            "NotResource": [
                "arn:aws:s3:::HRBucket/Payroll",
                "arn:aws:s3:::HRBucket/Payroll/*",
            ],
        },
    },
    {
        "Version": "2012-10-17",
        "Id": "sdfsdfsdf",
        "Statement": {
            "Effect": "Allow",
            "NotAction": "s3:ListBucket",
            "Resource": "arn:aws:s3:::example_bucket",
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "aaaaaadsfdsafsadfsadfaaaaa:ListBucket",
            "Resource": "arn:aws:s3:::example_bucket",
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3-s:ListBucket",
            "Resource": "arn:aws:s3:::example_bucket",
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3.s:ListBucket",
            "Resource": "arn:aws:s3:::example_bucket",
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "NotAction": "s3:ListBucket",
            "NotResource": "*",
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "sdf",
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
            },
            {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::example_bucket",
            },
        ],
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": "arn:aws:s3:::example_bucket",
            "Condition": {"DateGreaterThan": {"a": "01T"}},
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": "arn:aws:s3:::example_bucket",
            "Condition": {"x": {}, "y": {}},
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": "arn:aws:s3:::example_bucket",
            "Condition": {"StringEqualsIfExists": {"a": "asf"}},
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": "arn:aws:s3:::example_bucket",
            "Condition": {"ForAnyValue:StringEqualsIfExists": {"a": "asf"}},
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": "arn:aws:s3:::example_bucket",
            "Condition": {"DateLessThanEquals": {"a": "2019-07-01T13:20:15Z"}},
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": "arn:aws:s3:::example_bucket",
            "Condition": {
                "DateLessThanEquals": {"a": "2016-12-13T21:20:37.593194+00:00"}
            },
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": "arn:aws:s3:::example_bucket",
            "Condition": {"DateLessThanEquals": {"a": "2016-12-13t2:00:00.593194+23"}},
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "s3:ListBucket",
            "Resource": "arn:aws:s3:::example_bucket",
            "Condition": {"DateLessThan": {"a": "-292275054"}},
        },
    },
    {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "AllowViewAccountInfo",
                "Effect": "Allow",
                "Action": [
                    "iam:GetAccountPasswordPolicy",
                    "iam:GetAccountSummary",
                    "iam:ListVirtualMFADevices",
                ],
                "Resource": "*",
            },
            {
                "Sid": "AllowManageOwnPasswords",
                "Effect": "Allow",
                "Action": ["iam:ChangePassword", "iam:GetUser"],
                "Resource": "arn:aws:iam::*:user/${aws:username}",
            },
            {
                "Sid": "AllowManageOwnAccessKeys",
                "Effect": "Allow",
                "Action": [
                    "iam:CreateAccessKey",
                    "iam:DeleteAccessKey",
                    "iam:ListAccessKeys",
                    "iam:UpdateAccessKey",
                ],
                "Resource": "arn:aws:iam::*:user/${aws:username}",
            },
            {
                "Sid": "AllowManageOwnSigningCertificates",
                "Effect": "Allow",
                "Action": [
                    "iam:DeleteSigningCertificate",
                    "iam:ListSigningCertificates",
                    "iam:UpdateSigningCertificate",
                    "iam:UploadSigningCertificate",
                ],
                "Resource": "arn:aws:iam::*:user/${aws:username}",
            },
            {
                "Sid": "AllowManageOwnSSHPublicKeys",
                "Effect": "Allow",
                "Action": [
                    "iam:DeleteSSHPublicKey",
                    "iam:GetSSHPublicKey",
                    "iam:ListSSHPublicKeys",
                    "iam:UpdateSSHPublicKey",
                    "iam:UploadSSHPublicKey",
                ],
                "Resource": "arn:aws:iam::*:user/${aws:username}",
            },
            {
                "Sid": "AllowManageOwnGitCredentials",
                "Effect": "Allow",
                "Action": [
                    "iam:CreateServiceSpecificCredential",
                    "iam:DeleteServiceSpecificCredential",
                    "iam:ListServiceSpecificCredentials",
                    "iam:ResetServiceSpecificCredential",
                    "iam:UpdateServiceSpecificCredential",
                ],
                "Resource": "arn:aws:iam::*:user/${aws:username}",
            },
            {
                "Sid": "AllowManageOwnVirtualMFADevice",
                "Effect": "Allow",
                "Action": ["iam:CreateVirtualMFADevice", "iam:DeleteVirtualMFADevice"],
                "Resource": "arn:aws:iam::*:mfa/${aws:username}",
            },
            {
                "Sid": "AllowManageOwnUserMFA",
                "Effect": "Allow",
                "Action": [
                    "iam:DeactivateMFADevice",
                    "iam:EnableMFADevice",
                    "iam:ListMFADevices",
                    "iam:ResyncMFADevice",
                ],
                "Resource": "arn:aws:iam::*:user/${aws:username}",
            },
            {
                "Sid": "DenyAllExceptListedIfNoMFA",
                "Effect": "Deny",
                "NotAction": [
                    "iam:CreateVirtualMFADevice",
                    "iam:EnableMFADevice",
                    "iam:GetUser",
                    "iam:ListMFADevices",
                    "iam:ListVirtualMFADevices",
                    "iam:ResyncMFADevice",
                    "sts:GetSessionToken",
                ],
                "Resource": "*",
                "Condition": {"BoolIfExists": {"aws:MultiFactorAuthPresent": "false"}},
            },
        ],
    },
    {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "ListAndDescribe",
                "Effect": "Allow",
                "Action": [
                    "dynamodb:List*",
                    "dynamodb:DescribeReservedCapacity*",
                    "dynamodb:DescribeLimits",
                    "dynamodb:DescribeTimeToLive",
                ],
                "Resource": "*",
            },
            {
                "Sid": "SpecificTable",
                "Effect": "Allow",
                "Action": [
                    "dynamodb:BatchGet*",
                    "dynamodb:DescribeStream",
                    "dynamodb:DescribeTable",
                    "dynamodb:Get*",
                    "dynamodb:Query",
                    "dynamodb:Scan",
                    "dynamodb:BatchWrite*",
                    "dynamodb:CreateTable",
                    "dynamodb:Delete*",
                    "dynamodb:Update*",
                    "dynamodb:PutItem",
                ],
                "Resource": "arn:aws:dynamodb:*:*:table/MyTable",
            },
        ],
    },
    {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Action": ["ec2:AttachVolume", "ec2:DetachVolume"],
                "Resource": ["arn:aws:ec2:*:*:volume/*", "arn:aws:ec2:*:*:instance/*"],
                "Condition": {
                    "ArnEquals": {
                        "ec2:SourceInstanceARN": "arn:aws:ec2:*:*:instance/instance-id"
                    }
                },
            }
        ],
    },
    {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Action": ["ec2:AttachVolume", "ec2:DetachVolume"],
                "Resource": "arn:aws:ec2:*:*:instance/*",
                "Condition": {
                    "StringEquals": {"ec2:ResourceTag/Department": "Development"}
                },
            },
            {
                "Effect": "Allow",
                "Action": ["ec2:AttachVolume", "ec2:DetachVolume"],
                "Resource": "arn:aws:ec2:*:*:volume/*",
                "Condition": {
                    "StringEquals": {"ec2:ResourceTag/VolumeUser": "${aws:username}"}
                },
            },
        ],
    },
    {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "StartStopIfTags",
                "Effect": "Allow",
                "Action": [
                    "ec2:StartInstances",
                    "ec2:StopInstances",
                    "ec2:DescribeTags",
                ],
                "Resource": "arn:aws:ec2:region:account-id:instance/*",
                "Condition": {
                    "StringEquals": {
                        "ec2:ResourceTag/Project": "DataAnalytics",
                        "aws:PrincipalTag/Department": "Data",
                    }
                },
            }
        ],
    },
    {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "ListYourObjects",
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": ["arn:aws:s3:::bucket-name"],
                "Condition": {
                    "StringLike": {
                        "s3:prefix": [
                            "cognito/application-name/${cognito-identity.amazonaws.com:sub}"
                        ]
                    }
                },
            },
            {
                "Sid": "ReadWriteDeleteYourObjects",
                "Effect": "Allow",
                "Action": ["s3:GetObject", "s3:PutObject", "s3:DeleteObject"],
                "Resource": [
                    "arn:aws:s3:::bucket-name/cognito/application-name/${cognito-identity.amazonaws.com:sub}",
                    "arn:aws:s3:::bucket-name/cognito/application-name/${cognito-identity.amazonaws.com:sub}/*",
                ],
            },
        ],
    },
    {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Action": ["s3:ListAllMyBuckets", "s3:GetBucketLocation"],
                "Resource": "*",
            },
            {
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": "arn:aws:s3:::bucket-name",
                "Condition": {
                    "StringLike": {"s3:prefix": ["", "home/", "home/${aws:userid}/*"]}
                },
            },
            {
                "Effect": "Allow",
                "Action": "s3:*",
                "Resource": [
                    "arn:aws:s3:::bucket-name/home/${aws:userid}",
                    "arn:aws:s3:::bucket-name/home/${aws:userid}/*",
                ],
            },
        ],
    },
    {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "ConsoleAccess",
                "Effect": "Allow",
                "Action": [
                    "s3:GetAccountPublicAccessBlock",
                    "s3:GetBucketAcl",
                    "s3:GetBucketLocation",
                    "s3:GetBucketPolicyStatus",
                    "s3:GetBucketPublicAccessBlock",
                    "s3:ListAllMyBuckets",
                ],
                "Resource": "*",
            },
            {
                "Sid": "ListObjectsInBucket",
                "Effect": "Allow",
                "Action": "s3:ListBucket",
                "Resource": ["arn:aws:s3:::bucket-name"],
            },
            {
                "Sid": "AllObjectActions",
                "Effect": "Allow",
                "Action": "s3:*Object",
                "Resource": ["arn:aws:s3:::bucket-name/*"],
            },
        ],
    },
    {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "AllowViewAccountInfo",
                "Effect": "Allow",
                "Action": ["iam:GetAccountPasswordPolicy", "iam:GetAccountSummary"],
                "Resource": "*",
            },
            {
                "Sid": "AllowManageOwnPasswords",
                "Effect": "Allow",
                "Action": ["iam:ChangePassword", "iam:GetUser"],
                "Resource": "arn:aws:iam::*:user/${aws:username}",
            },
            {
                "Sid": "AllowManageOwnAccessKeys",
                "Effect": "Allow",
                "Action": [
                    "iam:CreateAccessKey",
                    "iam:DeleteAccessKey",
                    "iam:ListAccessKeys",
                    "iam:UpdateAccessKey",
                ],
                "Resource": "arn:aws:iam::*:user/${aws:username}",
            },
            {
                "Sid": "AllowManageOwnSigningCertificates",
                "Effect": "Allow",
                "Action": [
                    "iam:DeleteSigningCertificate",
                    "iam:ListSigningCertificates",
                    "iam:UpdateSigningCertificate",
                    "iam:UploadSigningCertificate",
                ],
                "Resource": "arn:aws:iam::*:user/${aws:username}",
            },
            {
                "Sid": "AllowManageOwnSSHPublicKeys",
                "Effect": "Allow",
                "Action": [
                    "iam:DeleteSSHPublicKey",
                    "iam:GetSSHPublicKey",
                    "iam:ListSSHPublicKeys",
                    "iam:UpdateSSHPublicKey",
                    "iam:UploadSSHPublicKey",
                ],
                "Resource": "arn:aws:iam::*:user/${aws:username}",
            },
            {
                "Sid": "AllowManageOwnGitCredentials",
                "Effect": "Allow",
                "Action": [
                    "iam:CreateServiceSpecificCredential",
                    "iam:DeleteServiceSpecificCredential",
                    "iam:ListServiceSpecificCredentials",
                    "iam:ResetServiceSpecificCredential",
                    "iam:UpdateServiceSpecificCredential",
                ],
                "Resource": "arn:aws:iam::*:user/${aws:username}",
            },
        ],
    },
    {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Action": "ec2:*",
                "Resource": "*",
                "Effect": "Allow",
                "Condition": {"StringEquals": {"ec2:Region": "region"}},
            }
        ],
    },
    {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Action": "rds:*",
                "Resource": ["arn:aws:rds:region:*:*"],
            },
            {"Effect": "Allow", "Action": ["rds:Describe*"], "Resource": ["*"]},
        ],
    },
    {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "",
                "Effect": "Allow",
                "Action": "rds:*",
                "Resource": ["arn:aws:rds:region:*:*"],
            },
            {
                "Sid": "",
                "Effect": "Allow",
                "Action": ["rds:Describe*"],
                "Resource": ["*"],
            },
        ],
    },
    {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "",
                "Effect": "Allow",
                "Action": ["s3:*"],
                "Resource": [
                    "arn:aws:s3:us-west-2:123456789012:accesspoint/my-access-point",
                    "arn:aws:s3:us-west-2:123456789012:accesspoint/my-access-point/object/*",
                ],
            },
        ],
    },
]


@pytest.mark.parametrize("invalid_policy_document", invalid_policy_document_test_cases)
@mock_aws
def test_create_policy_with_invalid_policy_document(invalid_policy_document):
    conn = boto3.client("iam", region_name="us-east-1")
    with pytest.raises(ClientError) as ex:
        conn.create_policy(
            PolicyName="TestCreatePolicy",
            PolicyDocument=json.dumps(invalid_policy_document["document"]),
        )
    resp = ex.value.response
    assert resp["Error"]["Code"] == "MalformedPolicyDocument"
    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert resp["Error"]["Message"] == invalid_policy_document["error_message"]


@pytest.mark.parametrize("valid_policy_document", valid_policy_documents)
@mock_aws
def test_create_policy_with_valid_policy_document(valid_policy_document):
    conn = boto3.client("iam", region_name="us-east-1")
    conn.create_policy(
        PolicyName="TestCreatePolicy", PolicyDocument=json.dumps(valid_policy_document)
    )



"""SESV2Backend class with methods for supported APIs."""

from typing import Any, Dict, List

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.core.utils import iso_8601_datetime_with_milliseconds
from moto.utilities.paginator import paginate

from ..ses.models import ConfigurationSet, Message, RawMessage, ses_backends
from .exceptions import NotFoundException

PAGINATION_MODEL = {
    "list_dedicated_ip_pools": {
        "input_token": "next_token",
        "limit_key": "page_size",
        "limit_default": 100,
        "unique_attribute": ["pool_name"],
    },
}


class Contact(BaseModel):
    def __init__(
        self,
        contact_list_name: str,
        email_address: str,
        topic_preferences: List[Dict[str, str]],
        unsubscribe_all: bool,
    ) -> None:
        self.contact_list_name = contact_list_name
        self.email_address = email_address
        self.topic_default_preferences: List[Dict[str, str]] = []
        self.topic_preferences = topic_preferences
        self.unsubscribe_all = unsubscribe_all
        self.created_timestamp = iso_8601_datetime_with_milliseconds()
        self.last_updated_timestamp = iso_8601_datetime_with_milliseconds()

    @property
    def response_object(self) -> Dict[str, Any]:  # type: ignore[misc]
        return {
            "ContactListName": self.contact_list_name,
            "EmailAddress": self.email_address,
            "TopicDefaultPreferences": self.topic_default_preferences,
            "TopicPreferences": self.topic_preferences,
            "UnsubscribeAll": self.unsubscribe_all,
            "CreatedTimestamp": self.created_timestamp,
            "LastUpdatedTimestamp": self.last_updated_timestamp,
        }


class ContactList(BaseModel):
    def __init__(
        self,
        contact_list_name: str,
        description: str,
        topics: List[Dict[str, str]],
    ) -> None:
        self.contact_list_name = contact_list_name
        self.description = description
        self.topics = topics
        self.created_timestamp = iso_8601_datetime_with_milliseconds()
        self.last_updated_timestamp = iso_8601_datetime_with_milliseconds()
        self.contacts: Dict[str, Contact] = {}

    def create_contact(self, contact_list_name: str, params: Dict[str, Any]) -> None:
        email_address = params["EmailAddress"]
        topic_preferences = (
            [] if "TopicPreferences" not in params else params["TopicPreferences"]
        )
        unsubscribe_all = (
            False if "UnsubscribeAll" not in params else params["UnsubscribeAll"]
        )
        new_contact = Contact(
            contact_list_name, email_address, topic_preferences, unsubscribe_all
        )
        self.contacts[email_address] = new_contact

    def list_contacts(self) -> List[Contact]:
        return self.contacts.values()  # type: ignore[return-value]

    def get_contact(self, email: str) -> Contact:
        if email in self.contacts:
            return self.contacts[email]
        else:
            raise NotFoundException(f"{email} doesn't exist in List.")

    def delete_contact(self, email: str) -> None:
        # delete if contact exists, otherwise get_contact will throw appropriate exception
        if self.get_contact(email):
            del self.contacts[email]

    @property
    def response_object(self) -> Dict[str, Any]:  # type: ignore[misc]
        return {
            "ContactListName": self.contact_list_name,
            "Description": self.description,
            "Topics": self.topics,
            "CreatedTimestamp": self.created_timestamp,
            "LastUpdatedTimestamp": self.last_updated_timestamp,
        }


class DedicatedIpPool(BaseModel):
    def __init__(
        self, pool_name: str, scaling_mode: str, tags: List[Dict[str, str]]
    ) -> None:
        self.pool_name = pool_name
        self.scaling_mode = scaling_mode
        self.tags = tags

    def to_dict(self) -> Dict[str, Any]:
        return {
            "PoolName": self.pool_name,
            "Tags": self.tags,
            "ScalingMode": self.scaling_mode,
        }


class SESV2Backend(BaseBackend):
    """Implementation of SESV2 APIs, piggy back on v1 SES"""

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.contacts: Dict[str, Contact] = {}
        self.contacts_lists: Dict[str, ContactList] = {}
        self.v1_backend = ses_backends[self.account_id][self.region_name]
        self.dedicated_ip_pools: Dict[str, DedicatedIpPool] = {}

    def create_contact_list(self, params: Dict[str, Any]) -> None:
        name = params["ContactListName"]
        description = params.get("Description")
        topics = [] if "Topics" not in params else params["Topics"]
        new_list = ContactList(name, str(description), topics)
        self.contacts_lists[name] = new_list

    def get_contact_list(self, contact_list_name: str) -> ContactList:
        if contact_list_name in self.contacts_lists:
            return self.contacts_lists[contact_list_name]
        else:
            raise NotFoundException(
                f"List with name: {contact_list_name} doesn't exist."
            )

    def list_contact_lists(self) -> List[ContactList]:
        return self.contacts_lists.values()  # type: ignore[return-value]

    def delete_contact_list(self, name: str) -> None:
        if name in self.contacts_lists:
            del self.contacts_lists[name]
        else:
            raise NotFoundException(f"List with name: {name} doesn't exist")

    def create_contact(self, contact_list_name: str, params: Dict[str, Any]) -> None:
        contact_list = self.get_contact_list(contact_list_name)
        contact_list.create_contact(contact_list_name, params)
        return

    def get_contact(self, email: str, contact_list_name: str) -> Contact:
        contact_list = self.get_contact_list(contact_list_name)
        contact = contact_list.get_contact(email)
        return contact

    def list_contacts(self, contact_list_name: str) -> List[Contact]:
        contact_list = self.get_contact_list(contact_list_name)
        contacts = contact_list.list_contacts()
        return contacts

    def delete_contact(self, email: str, contact_list_name: str) -> None:
        contact_list = self.get_contact_list(contact_list_name)
        contact_list.delete_contact(email)
        return

    def send_email(
        self, source: str, destinations: Dict[str, List[str]], subject: str, body: str
    ) -> Message:
        message = self.v1_backend.send_email(
            source=source,
            destinations=destinations,
            subject=subject,
            body=body,
        )
        return message

    def send_raw_email(
        self, source: str, destinations: List[str], raw_data: str
    ) -> RawMessage:
        message = self.v1_backend.send_raw_email(
            source=source, destinations=destinations, raw_data=raw_data
        )
        return message

    def create_configuration_set(
        self,
        configuration_set_name: str,
        tracking_options: Dict[str, str],
        delivery_options: Dict[str, Any],
        reputation_options: Dict[str, Any],
        sending_options: Dict[str, bool],
        tags: List[Dict[str, str]],
        suppression_options: Dict[str, List[str]],
        vdm_options: Dict[str, Dict[str, str]],
    ) -> None:
        self.v1_backend.create_configuration_set_v2(
            configuration_set_name=configuration_set_name,
            tracking_options=tracking_options,
            delivery_options=delivery_options,
            reputation_options=reputation_options,
            sending_options=sending_options,
            tags=tags,
            suppression_options=suppression_options,
            vdm_options=vdm_options,
        )

    def delete_configuration_set(self, configuration_set_name: str) -> None:
        self.v1_backend.delete_configuration_set(configuration_set_name)

    def get_configuration_set(self, configuration_set_name: str) -> ConfigurationSet:
        config_set = self.v1_backend.describe_configuration_set(
            configuration_set_name=configuration_set_name
        )
        return config_set

    def list_configuration_sets(self, next_token: str, page_size: int) -> List[str]:
        return self.v1_backend.list_configuration_sets(
            next_token=next_token, max_items=page_size
        )

    def create_dedicated_ip_pool(
        self, pool_name: str, tags: List[Dict[str, str]], scaling_mode: str
    ) -> None:
        if pool_name not in self.dedicated_ip_pools:
            new_pool = DedicatedIpPool(
                pool_name=pool_name, tags=tags, scaling_mode=scaling_mode
            )
            self.dedicated_ip_pools[pool_name] = new_pool

    def delete_dedicated_ip_pool(self, pool_name: str) -> None:
        self.dedicated_ip_pools.pop(pool_name)

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_dedicated_ip_pools(self) -> List[str]:
        return list(self.dedicated_ip_pools.keys())

    def get_dedicated_ip_pool(self, pool_name: str) -> DedicatedIpPool:
        if not self.dedicated_ip_pools.get(pool_name, None):
            raise NotFoundException(pool_name)
        return self.dedicated_ip_pools[pool_name]


sesv2_backends = BackendDict(SESV2Backend, "sesv2")



"""route53domains base URL and path."""

from .responses import Route53DomainsResponse

url_bases = [
    r"https?://route53domains\.(.+)\.amazonaws\.com",
]


url_paths = {
    "{0}/$": Route53DomainsResponse.dispatch,
}



import datetime

import boto3

from moto import mock_aws


@mock_aws
def test_create_node_from_template_job() -> None:
    # Given
    panorama_client = boto3.client("panorama", "us-east-1")

    # When
    response = panorama_client.create_node_from_template_job(
        JobTags=[
            {"ResourceType": "PACKAGE", "Tags": {"key": "value"}},
        ],
        NodeDescription="a description",
        NodeName="a-camera-node",
        OutputPackageName="an-output-package-name",
        OutputPackageVersion="1.0",
        TemplateParameters={
            "Username": "a-user-name",
            "Password": "a-password",
            "StreamUrl": "rtsp://",
        },
        TemplateType="RTSP_CAMERA_STREAM",
    )

    # Then
    assert response["ResponseMetadata"]["HTTPStatusCode"] == 200
    assert "JobId" in response
    assert isinstance(response["JobId"], str)


@mock_aws
def test_describe_node_from_template_job() -> None:
    # Given
    panorama_client = boto3.client("panorama", "us-east-1")

    # When
    response = panorama_client.create_node_from_template_job(
        JobTags=[
            {"ResourceType": "PACKAGE", "Tags": {"key": "value"}},
        ],
        NodeDescription="a description",
        NodeName="not-a-camera-node",
        OutputPackageName="not-an-output-package-name",
        OutputPackageVersion="1.0",
        TemplateParameters={
            "Username": "not-a-user-name",
            "Password": "not-a-password",
            "StreamUrl": "rtsp://192.168.0.201/1",
        },
        TemplateType="RTSP_CAMERA_STREAM",
    )
    describe_response = panorama_client.describe_node_from_template_job(
        JobId=response["JobId"]
    )

    # Then
    assert response["ResponseMetadata"]["HTTPStatusCode"] == 200
    assert "JobId" in response
    assert isinstance(response["JobId"], str)
    assert isinstance(describe_response["CreatedTime"], datetime.datetime)
    assert describe_response["JobId"] == response["JobId"]
    assert describe_response["JobTags"] == [
        {"ResourceType": "PACKAGE", "Tags": {"key": "value"}}
    ]
    assert isinstance(describe_response["LastUpdatedTime"], datetime.datetime)
    # Weird, seems to be a bug in panorama
    assert describe_response["NodeDescription"] == "not-a-camera-node"
    assert describe_response["NodeName"] == "not-a-camera-node"
    assert describe_response["OutputPackageName"] == "not-an-output-package-name"
    assert describe_response["OutputPackageVersion"] == "1.0"
    assert describe_response["Status"] == "PENDING"
    assert describe_response["TemplateParameters"] == {
        "Username": "SAVED_AS_SECRET",
        "Password": "SAVED_AS_SECRET",
        "StreamUrl": "rtsp://192.168.0.201/1",
    }
    assert describe_response["TemplateType"] == "RTSP_CAMERA_STREAM"


@mock_aws
def test_list_nodes() -> None:
    # Given
    panorama_client = boto3.client("panorama", "us-east-1")
    panorama_client.create_node_from_template_job(
        JobTags=[
            {"ResourceType": "PACKAGE", "Tags": {"key": "value"}},
        ],
        NodeDescription="a description",
        NodeName="not-a-camera-node",
        OutputPackageName="not-an-output-package-name",
        OutputPackageVersion="1.0",
        TemplateParameters={
            "Username": "not-a-user-name",
            "Password": "not-a-password",
            "StreamUrl": "rtsp://",
        },
        TemplateType="RTSP_CAMERA_STREAM",
    )
    panorama_client.create_node_from_template_job(
        JobTags=[
            {"ResourceType": "PACKAGE", "Tags": {"key": "value"}},
        ],
        NodeDescription="a description",
        NodeName="not-another-camera-node",
        OutputPackageName="not-another-output-package-name",
        OutputPackageVersion="1.0",
        TemplateParameters={
            "Username": "not-another-user-name",
            "Password": "not-another-password",
            "StreamUrl": "rtsp://",
        },
        TemplateType="RTSP_CAMERA_STREAM",
    )
    given_category = "MEDIA_SOURCE"

    # When
    response_page_1 = panorama_client.list_nodes(Category=given_category, MaxResults=1)
    response_page_2 = panorama_client.list_nodes(
        Category=given_category, MaxResults=1, NextToken=response_page_1["NextToken"]
    )

    # Then
    assert response_page_1["ResponseMetadata"]["HTTPStatusCode"] == 200
    assert "Nodes" in response_page_1
    assert isinstance(response_page_1["Nodes"], list)
    assert len(response_page_1["Nodes"]) == 1
    assert "NextToken" in response_page_1
    assert isinstance(response_page_1["NextToken"], str)
    node = response_page_1["Nodes"][0]
    assert node["Category"] == given_category
    assert isinstance(node["CreatedTime"], datetime.datetime)
    assert node["Description"] == "a description"
    assert node["Name"] == "not-a-camera-node"
    assert node["OwnerAccount"] == "123456789012"
    assert (
        node["PackageArn"]
        == "arn:aws:panorama:us-east-1:123456789012:package/package--Y6-ycqU-ogoBTHuibMzGg"
    )
    assert node["PackageId"] == "package--Y6-ycqU-ogoBTHuibMzGg"
    assert node["PackageName"] == "not-an-output-package-name"
    assert node["PackageVersion"] == "1.0"
    assert (
        node["PatchVersion"]
        == "y6ycquogobthuibmzggy6ycquogobthuibmzggy6ycquogobthuibmzggy6ycquo"
    )

    assert response_page_2["ResponseMetadata"]["HTTPStatusCode"] == 200
    assert "Nodes" in response_page_2
    assert isinstance(response_page_2["Nodes"], list)
    assert len(response_page_2["Nodes"]) == 1
    assert "NextToken" not in response_page_2



from typing import Any, List, Optional, Tuple

from botocore.awsrequest import AWSPreparedRequest
from werkzeug.wrappers import Request

from moto import settings
from moto.core.common_types import TYPE_RESPONSE
from moto.core.responses import BaseResponse
from moto.ec2.models import ec2_backends

from .exceptions import DBParameterGroupNotFoundError, RDSClientError
from .models import RDSBackend, rds_backends
from .parser import QueryParser, XFormedDict
from .serialize import QuerySerializer
from .utils import get_service_model


def normalize_request(request: AWSPreparedRequest) -> Request:
    from urllib.parse import urlparse

    parsed_url = urlparse(request.url)
    normalized_request = Request.from_values(
        method=request.method,
        base_url=f"{parsed_url.scheme}://{parsed_url.netloc}",
        path=parsed_url.path,
        query_string=parsed_url.query,
        data=request.body,
        headers=[(k, v) for k, v in request.headers.items()],
    )
    return normalized_request


class RDSResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="rds")

    @property
    def backend(self) -> RDSBackend:
        return rds_backends[self.current_account][self.region]

    @property
    def global_backend(self) -> RDSBackend:
        """Return backend instance of the region that stores Global Clusters"""
        return rds_backends[self.current_account]["us-east-1"]

    def _dispatch(self, request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:
        self.setup_class(request, full_url, headers)

        if isinstance(request, AWSPreparedRequest):
            request = normalize_request(request)

        self.action = request.values["Action"]

        service_model = get_service_model(self.service_name)
        self.operation_model = service_model.operation_model(self.action)

        parser = QueryParser(map_type=XFormedDict)  # type: ignore[no-untyped-call]
        self.parameters = parser.parse(
            {"query_params": request.values},
            self.operation_model,  # type: ignore[no-untyped-call]
        )

        self.serializer = QuerySerializer(
            self.operation_model,
            {"request_id": "request-id"},
            pretty_print=settings.PRETTIFY_RESPONSES,
        )
        try:
            response = self.call_action()
        except RDSClientError as e:
            response = self.serialize(e)
        return response

    def serialize(self, result: Any) -> TYPE_RESPONSE:
        serialized = self.serializer.serialize_to_response(result)
        return serialized["status_code"], serialized["headers"], serialized["body"]

    def create_db_instance(self) -> TYPE_RESPONSE:
        db_kwargs = self.parameters
        database = self.backend.create_db_instance(db_kwargs)
        result = {"DBInstance": database}
        return self.serialize(result)

    def create_db_instance_read_replica(self) -> TYPE_RESPONSE:
        db_kwargs = self.parameters
        database = self.backend.create_db_instance_read_replica(db_kwargs)
        result = {"DBInstance": database}
        return self.serialize(result)

    def describe_db_instances(self) -> TYPE_RESPONSE:
        db_instance_identifier = self.parameters.get("DBInstanceIdentifier")
        filters = self.parameters.get("Filters", [])
        filter_dict = {f["Name"]: f["Values"] for f in filters}
        all_instances = list(
            self.backend.describe_db_instances(
                db_instance_identifier, filters=filter_dict
            )
        )
        instances_resp, next_marker = self._paginate(all_instances)
        result = {
            "DBInstances": instances_resp,
            "Marker": next_marker,
        }
        return self.serialize(result)

    def modify_db_instance(self) -> TYPE_RESPONSE:
        db_instance_identifier = self.parameters.get("DBInstanceIdentifier")
        db_kwargs = self.parameters
        # This is a hack because the backend code expects the parameter to be
        # lowercased before validation is performed.  We need to move the
        # lowercasing to a backend setter (in one place) and then do the validation.
        if "PreferredMaintenanceWindow" in db_kwargs:
            db_kwargs["PreferredMaintenanceWindow"] = db_kwargs[
                "PreferredMaintenanceWindow"
            ].lower()
        new_db_instance_identifier = self.parameters.get("NewDBInstanceIdentifier")
        if new_db_instance_identifier:
            db_kwargs["new_db_instance_identifier"] = new_db_instance_identifier
        database = self.backend.modify_db_instance(db_instance_identifier, db_kwargs)
        result = {"DBInstance": database}
        return self.serialize(result)

    def delete_db_instance(self) -> TYPE_RESPONSE:
        db_snapshot_name = self.parameters.get("FinalDBSnapshotIdentifier")
        if db_snapshot_name is not None:
            self.backend.validate_db_snapshot_identifier(
                db_snapshot_name, parameter_name="FinalDBSnapshotIdentifier"
            )
        database = self.backend.delete_db_instance(**self.parameters)
        result = {"DBInstance": database}
        return self.serialize(result)

    def reboot_db_instance(self) -> TYPE_RESPONSE:
        db_instance_identifier = self.parameters.get("DBInstanceIdentifier")
        database = self.backend.reboot_db_instance(db_instance_identifier)
        result = {"DBInstance": database}
        return self.serialize(result)

    def create_db_snapshot(self) -> TYPE_RESPONSE:
        self.backend.validate_db_snapshot_identifier(
            self.parameters["DBSnapshotIdentifier"],
            parameter_name="DBSnapshotIdentifier",
        )
        snapshot = self.backend.create_db_snapshot(**self.parameters)
        result = {"DBSnapshot": snapshot}
        return self.serialize(result)

    def copy_db_snapshot(self) -> TYPE_RESPONSE:
        target_snapshot_identifier = self.parameters.get("TargetDBSnapshotIdentifier")
        self.backend.validate_db_snapshot_identifier(
            target_snapshot_identifier, parameter_name="TargetDBSnapshotIdentifier"
        )
        snapshot = self.backend.copy_db_snapshot(**self.parameters)
        result = {"DBSnapshot": snapshot}
        return self.serialize(result)

    def describe_db_snapshots(self) -> TYPE_RESPONSE:
        db_instance_identifier = self.parameters.get("DBInstanceIdentifier")
        db_snapshot_identifier = self.parameters.get("DBSnapshotIdentifier")
        snapshot_type = self.parameters.get("SnapshotType")
        filters = self.parameters.get("Filters", [])
        filter_dict = {f["Name"]: f["Values"] for f in filters}
        snapshots = self.backend.describe_db_snapshots(
            db_instance_identifier, db_snapshot_identifier, snapshot_type, filter_dict
        )
        result = {"DBSnapshots": snapshots}
        return self.serialize(result)

    def promote_read_replica(self) -> TYPE_RESPONSE:
        db_kwargs = self.parameters
        database = self.backend.promote_read_replica(db_kwargs)
        result = {"DBInstance": database}
        return self.serialize(result)

    def delete_db_snapshot(self) -> TYPE_RESPONSE:
        db_snapshot_identifier = self.parameters.get("DBSnapshotIdentifier")
        snapshot = self.backend.delete_db_snapshot(db_snapshot_identifier)
        result = {"DBSnapshot": snapshot}
        return self.serialize(result)

    def restore_db_instance_from_db_snapshot(self) -> TYPE_RESPONSE:
        db_snapshot_identifier = self.parameters.get("DBSnapshotIdentifier")
        db_kwargs = self.parameters
        new_instance = self.backend.restore_db_instance_from_db_snapshot(
            db_snapshot_identifier, db_kwargs
        )
        result = {"DBInstance": new_instance}
        return self.serialize(result)

    def restore_db_instance_to_point_in_time(self) -> TYPE_RESPONSE:
        source_db_identifier = self.parameters.get("SourceDBInstanceIdentifier")
        target_db_identifier = self.parameters.get("TargetDBInstanceIdentifier")

        db_kwargs = self.parameters
        new_instance = self.backend.restore_db_instance_to_point_in_time(
            source_db_identifier, target_db_identifier, db_kwargs
        )
        result = {"DBInstance": new_instance}
        return self.serialize(result)

    def restore_db_cluster_to_point_in_time(self) -> TYPE_RESPONSE:
        cluster = self.backend.restore_db_cluster_to_point_in_time(**self.parameters)
        result = {"DBCluster": cluster}
        return self.serialize(result)

    def failover_db_cluster(self) -> TYPE_RESPONSE:
        cluster = self.backend.failover_db_cluster(**self.parameters)
        result = {"DBCluster": cluster}
        return self.serialize(result)

    def list_tags_for_resource(self) -> TYPE_RESPONSE:
        arn = self.parameters.get("ResourceName")
        tags = self.backend.list_tags_for_resource(arn)
        result = {"TagList": tags}
        return self.serialize(result)

    def add_tags_to_resource(self) -> TYPE_RESPONSE:
        arn = self.parameters.get("ResourceName")
        tags = self.parameters.get("Tags", [])
        self.backend.add_tags_to_resource(arn, tags)
        return self.serialize({})

    def remove_tags_from_resource(self) -> TYPE_RESPONSE:
        arn = self.parameters.get("ResourceName")
        tag_keys = self.parameters.get("TagKeys")
        self.backend.remove_tags_from_resource(arn, tag_keys)
        return self.serialize({})

    def stop_db_instance(self) -> TYPE_RESPONSE:
        db_instance_identifier = self.parameters.get("DBInstanceIdentifier")
        db_snapshot_identifier = self.parameters.get("DBSnapshotIdentifier")
        if db_snapshot_identifier is not None:
            self.backend.validate_db_snapshot_identifier(
                db_snapshot_identifier, parameter_name="DBSnapshotIdentifier"
            )

        database = self.backend.stop_db_instance(
            db_instance_identifier, db_snapshot_identifier
        )
        result = {"DBInstance": database}
        return self.serialize(result)

    def start_db_instance(self) -> TYPE_RESPONSE:
        db_instance_identifier = self.parameters.get("DBInstanceIdentifier")
        database = self.backend.start_db_instance(db_instance_identifier)
        result = {"DBInstance": database}
        return self.serialize(result)

    def create_db_security_group(self) -> TYPE_RESPONSE:
        group_name = self.parameters.get("DBSecurityGroupName")
        description = self.parameters.get("DBSecurityGroupDescription")
        tags = self.parameters.get("Tags", [])
        security_group = self.backend.create_db_security_group(
            group_name, description, tags
        )
        result = {"DBSecurityGroup": security_group}
        return self.serialize(result)

    def describe_db_security_groups(self) -> TYPE_RESPONSE:
        security_group_name = self.parameters.get("DBSecurityGroupName")
        security_groups = self.backend.describe_security_groups(security_group_name)
        result = {"DBSecurityGroups": security_groups}
        return self.serialize(result)

    def delete_db_security_group(self) -> TYPE_RESPONSE:
        security_group_name = self.parameters.get("DBSecurityGroupName")
        security_group = self.backend.delete_security_group(security_group_name)
        result = {"DBSecurityGroup": security_group}
        return self.serialize(result)

    def authorize_db_security_group_ingress(self) -> TYPE_RESPONSE:
        security_group_name = self.parameters.get("DBSecurityGroupName")
        cidr_ip = self.parameters.get("CIDRIP")
        security_group = self.backend.authorize_security_group(
            security_group_name, cidr_ip
        )
        result = {"DBSecurityGroup": security_group}
        return self.serialize(result)

    def create_db_subnet_group(self) -> TYPE_RESPONSE:
        subnet_name = self.parameters.get("DBSubnetGroupName")
        description = self.parameters.get("DBSubnetGroupDescription")
        subnet_ids = self.parameters.get("SubnetIds", [])
        tags = self.parameters.get("Tags", [])
        subnets = [
            ec2_backends[self.current_account][self.region].get_subnet(subnet_id)
            for subnet_id in subnet_ids
        ]
        subnet_group = self.backend.create_subnet_group(
            subnet_name, description, subnets, tags
        )
        result = {"DBSubnetGroup": subnet_group}
        return self.serialize(result)

    def describe_db_subnet_groups(self) -> TYPE_RESPONSE:
        subnet_name = self.parameters.get("DBSubnetGroupName")
        subnet_groups = self.backend.describe_db_subnet_groups(subnet_name)
        result = {"DBSubnetGroups": subnet_groups}
        return self.serialize(result)

    def modify_db_subnet_group(self) -> TYPE_RESPONSE:
        subnet_name = self.parameters.get("DBSubnetGroupName")
        description = self.parameters.get("DBSubnetGroupDescription")
        subnet_ids = self.parameters.get("SubnetIds", [])
        subnets = [
            ec2_backends[self.current_account][self.region].get_subnet(subnet_id)
            for subnet_id in subnet_ids
        ]
        subnet_group = self.backend.modify_db_subnet_group(
            subnet_name, description, subnets
        )
        result = {"DBSubnetGroup": subnet_group}
        return self.serialize(result)

    def delete_db_subnet_group(self) -> TYPE_RESPONSE:
        subnet_name = self.parameters.get("DBSubnetGroupName")
        subnet_group = self.backend.delete_subnet_group(subnet_name)
        result = {"DBSubnetGroup": subnet_group}
        return self.serialize(result)

    def create_option_group(self) -> TYPE_RESPONSE:
        kwargs = self.parameters
        option_group = self.backend.create_option_group(kwargs)
        result = {"OptionGroup": option_group}
        return self.serialize(result)

    def delete_option_group(self) -> TYPE_RESPONSE:
        name = self.parameters["OptionGroupName"]
        option_group = self.backend.delete_option_group(name)
        result = {"OptionGroup": option_group}
        return self.serialize(result)

    def describe_option_groups(self) -> TYPE_RESPONSE:
        kwargs = self.parameters
        option_groups = self.backend.describe_option_groups(kwargs)
        option_groups, marker = self._paginate(option_groups)
        result = {
            "OptionGroupsList": option_groups,
            "Marker": marker,
        }
        return self.serialize(result)

    def describe_option_group_options(self) -> str:
        engine_name = self.parameters.get("EngineName")
        major_engine_version = self.parameters.get("MajorEngineVersion")
        return self.backend.describe_option_group_options(
            engine_name, major_engine_version
        )

    def modify_option_group(self) -> TYPE_RESPONSE:
        option_group_name = self.parameters.get("OptionGroupName")
        options_to_include = self.parameters.get("OptionsToInclude", [])
        options_to_remove = self.parameters.get("OptionsToRemove", [])
        option_group = self.backend.modify_option_group(
            option_group_name, options_to_include, options_to_remove
        )
        result = {"OptionGroup": option_group}
        return self.serialize(result)

    def create_db_parameter_group(self) -> TYPE_RESPONSE:
        kwargs = self.parameters
        db_parameter_group = self.backend.create_db_parameter_group(kwargs)
        result = {"DBParameterGroup": db_parameter_group}
        return self.serialize(result)

    def describe_db_parameter_groups(self) -> TYPE_RESPONSE:
        kwargs = self.parameters
        db_parameter_groups = self.backend.describe_db_parameter_groups(kwargs)
        db_parameter_groups, _ = self._paginate(db_parameter_groups)
        result = {"DBParameterGroups": db_parameter_groups}
        return self.serialize(result)

    def modify_db_parameter_group(self) -> TYPE_RESPONSE:
        db_parameter_group_name = self.parameters.get("DBParameterGroupName")
        param_list = self.parameters.get("Parameters", [])
        # Raw dict is stored on the backend, so we need the original PascalCase items.
        db_parameter_group_parameters = [
            dict(param.original_items()) for param in param_list
        ]
        db_parameter_group = self.backend.modify_db_parameter_group(
            db_parameter_group_name, db_parameter_group_parameters
        )
        result = {"DBParameterGroupName": db_parameter_group.name}
        return self.serialize(result)

    def describe_db_parameters(self) -> TYPE_RESPONSE:
        db_parameter_group_name = self.parameters.get("DBParameterGroupName")
        db_parameter_groups = self.backend.describe_db_parameter_groups(
            {"name": db_parameter_group_name}
        )
        if not db_parameter_groups:
            raise DBParameterGroupNotFoundError(db_parameter_group_name)
        parameters = db_parameter_groups[0].parameters.values()
        result = {"Parameters": parameters}
        return self.serialize(result)

    def delete_db_parameter_group(self) -> TYPE_RESPONSE:
        name = self.parameters["DBParameterGroupName"]
        db_parameter_group = self.backend.delete_db_parameter_group(name)
        return self.serialize(db_parameter_group)

    def describe_db_cluster_parameters(self) -> TYPE_RESPONSE:
        # TODO: This never worked at all...
        db_parameter_group_name = self.parameters.get("DBParameterGroupName")
        db_parameter_groups = self.backend.describe_db_cluster_parameters()
        if db_parameter_groups is None:
            raise DBParameterGroupNotFoundError(db_parameter_group_name)
        result = {"Parameters": db_parameter_groups}
        return self.serialize(result)

    def create_db_cluster(self) -> TYPE_RESPONSE:
        kwargs = self.parameters
        cluster = self.backend.create_db_cluster(kwargs)
        result = {"DBCluster": cluster}
        return self.serialize(result)

    def modify_db_cluster(self) -> TYPE_RESPONSE:
        kwargs = self.parameters
        cluster = self.backend.modify_db_cluster(kwargs)
        result = {"DBCluster": cluster}
        return self.serialize(result)

    def describe_db_clusters(self) -> TYPE_RESPONSE:
        _id = self.parameters.get("DBClusterIdentifier")
        filters = self.parameters.get("Filters", [])
        filter_dict = {f["Name"]: f["Values"] for f in filters}
        clusters = self.backend.describe_db_clusters(
            db_cluster_identifier=_id, filters=filter_dict
        )
        result = {"DBClusters": clusters}
        return self.serialize(result)

    def delete_db_cluster(self) -> TYPE_RESPONSE:
        _id = self.parameters.get("DBClusterIdentifier")
        snapshot_name = self.parameters.get("FinalDBSnapshotIdentifier")
        cluster = self.backend.delete_db_cluster(
            cluster_identifier=_id, snapshot_name=snapshot_name
        )
        result = {"DBCluster": cluster}
        return self.serialize(result)

    def start_db_cluster(self) -> TYPE_RESPONSE:
        _id = self.parameters.get("DBClusterIdentifier")
        cluster = self.backend.start_db_cluster(cluster_identifier=_id)
        result = {"DBCluster": cluster}
        return self.serialize(result)

    def stop_db_cluster(self) -> TYPE_RESPONSE:
        _id = self.parameters.get("DBClusterIdentifier")
        cluster = self.backend.stop_db_cluster(cluster_identifier=_id)
        result = {"DBCluster": cluster}
        return self.serialize(result)

    def create_db_cluster_snapshot(self) -> TYPE_RESPONSE:
        snapshot = self.backend.create_db_cluster_snapshot(**self.parameters)
        result = {"DBClusterSnapshot": snapshot}
        return self.serialize(result)

    def copy_db_cluster_snapshot(self) -> TYPE_RESPONSE:
        snapshot = self.backend.copy_db_cluster_snapshot(**self.parameters)
        result = {"DBClusterSnapshot": snapshot}
        return self.serialize(result)

    def describe_db_cluster_snapshots(self) -> TYPE_RESPONSE:
        db_cluster_identifier = self.parameters.get("DBClusterIdentifier")
        db_snapshot_identifier = self.parameters.get("DBClusterSnapshotIdentifier")
        snapshot_type = self.parameters.get("SnapshotType")
        filters = self.parameters.get("Filters", [])
        filter_dict = {f["Name"]: f["Values"] for f in filters}
        snapshots = self.backend.describe_db_cluster_snapshots(
            db_cluster_identifier, db_snapshot_identifier, snapshot_type, filter_dict
        )
        results = {"DBClusterSnapshots": snapshots}
        return self.serialize(results)

    def delete_db_cluster_snapshot(self) -> TYPE_RESPONSE:
        db_snapshot_identifier = self.parameters["DBClusterSnapshotIdentifier"]
        snapshot = self.backend.delete_db_cluster_snapshot(db_snapshot_identifier)
        result = {"DBClusterSnapshot": snapshot}
        return self.serialize(result)

    def restore_db_cluster_from_snapshot(self) -> TYPE_RESPONSE:
        db_snapshot_identifier = self.parameters.get("SnapshotIdentifier")
        db_kwargs = self.parameters
        new_cluster = self.backend.restore_db_cluster_from_snapshot(
            db_snapshot_identifier, db_kwargs
        )
        result = {"DBCluster": new_cluster}
        return self.serialize(result)

    def start_export_task(self) -> TYPE_RESPONSE:
        kwargs = self.parameters
        export_task = self.backend.start_export_task(kwargs)
        return self.serialize(export_task)

    def cancel_export_task(self) -> TYPE_RESPONSE:
        export_task_identifier = self.parameters.get("ExportTaskIdentifier")
        export_task = self.backend.cancel_export_task(export_task_identifier)
        return self.serialize(export_task)

    def describe_export_tasks(self) -> TYPE_RESPONSE:
        export_task_identifier = self.parameters.get("ExportTaskIdentifier")
        tasks = self.backend.describe_export_tasks(export_task_identifier)
        result = {"ExportTasks": tasks}
        return self.serialize(result)

    def create_event_subscription(self) -> TYPE_RESPONSE:
        kwargs = self.parameters
        subscription = self.backend.create_event_subscription(kwargs)
        result = {"EventSubscription": subscription}
        return self.serialize(result)

    def delete_event_subscription(self) -> TYPE_RESPONSE:
        subscription_name = self.parameters.get("SubscriptionName")
        subscription = self.backend.delete_event_subscription(subscription_name)
        result = {"EventSubscription": subscription}
        return self.serialize(result)

    def describe_event_subscriptions(self) -> TYPE_RESPONSE:
        subscription_name = self.parameters.get("SubscriptionName")
        subscriptions = self.backend.describe_event_subscriptions(subscription_name)
        result = {"EventSubscriptionsList": subscriptions}
        return self.serialize(result)

    def describe_orderable_db_instance_options(self) -> TYPE_RESPONSE:
        engine = self.parameters.get("Engine")
        engine_version = self.parameters.get("EngineVersion")
        options = self.backend.describe_orderable_db_instance_options(
            engine, engine_version
        )
        result = {"OrderableDBInstanceOptions": options}
        return self.serialize(result)

    def describe_global_clusters(self) -> TYPE_RESPONSE:
        clusters = self.global_backend.describe_global_clusters()
        result = {"GlobalClusters": clusters}
        return self.serialize(result)

    def create_global_cluster(self) -> TYPE_RESPONSE:
        params = self.parameters
        cluster = self.global_backend.create_global_cluster(
            global_cluster_identifier=params["GlobalClusterIdentifier"],
            source_db_cluster_identifier=params.get("SourceDBClusterIdentifier"),
            engine=params.get("Engine"),
            engine_version=params.get("EngineVersion"),
            storage_encrypted=params.get("StorageEncrypted"),
            deletion_protection=params.get("DeletionProtection"),
        )
        result = {"GlobalCluster": cluster}
        return self.serialize(result)

    def delete_global_cluster(self) -> TYPE_RESPONSE:
        params = self.parameters
        cluster = self.global_backend.delete_global_cluster(
            global_cluster_identifier=params["GlobalClusterIdentifier"],
        )
        result = {"GlobalCluster": cluster}
        return self.serialize(result)

    def remove_from_global_cluster(self) -> TYPE_RESPONSE:
        params = self.parameters
        global_cluster = self.backend.remove_from_global_cluster(
            global_cluster_identifier=params["GlobalClusterIdentifier"],
            db_cluster_identifier=params["DbClusterIdentifier"],
        )
        result = {"GlobalCluster": global_cluster}
        return self.serialize(result)

    def create_db_cluster_parameter_group(self) -> TYPE_RESPONSE:
        group_name = self.parameters.get("DBClusterParameterGroupName")
        family = self.parameters.get("DBParameterGroupFamily")
        desc = self.parameters.get("Description")
        db_cluster_parameter_group = self.backend.create_db_cluster_parameter_group(
            group_name=group_name,
            family=family,
            description=desc,
        )
        result = {"DBClusterParameterGroup": db_cluster_parameter_group}
        return self.serialize(result)

    def describe_db_cluster_parameter_groups(self) -> TYPE_RESPONSE:
        group_name = self.parameters.get("DBClusterParameterGroupName")
        db_parameter_groups = self.backend.describe_db_cluster_parameter_groups(
            group_name=group_name,
        )
        result = {"DBClusterParameterGroups": db_parameter_groups}
        return self.serialize(result)

    def delete_db_cluster_parameter_group(self) -> TYPE_RESPONSE:
        group_name = self.parameters.get("DBClusterParameterGroupName")
        self.backend.delete_db_cluster_parameter_group(
            group_name=group_name,
        )
        return self.serialize({})

    def promote_read_replica_db_cluster(self) -> TYPE_RESPONSE:
        db_cluster_identifier = self.parameters.get("DBClusterIdentifier")
        cluster = self.backend.promote_read_replica_db_cluster(db_cluster_identifier)
        result = {"DBCluster": cluster}
        return self.serialize(result)

    def describe_db_snapshot_attributes(self) -> TYPE_RESPONSE:
        params = self.parameters
        db_snapshot_identifier = params["DBSnapshotIdentifier"]
        db_snapshot_attributes_result = self.backend.describe_db_snapshot_attributes(
            db_snapshot_identifier=db_snapshot_identifier,
        )
        result = {
            "DBSnapshotAttributesResult": {
                "DBSnapshotIdentifier": db_snapshot_identifier,
                "DBSnapshotAttributes": [
                    {"AttributeName": name, "AttributeValues": values}
                    for name, values in db_snapshot_attributes_result.items()
                ],
            }
        }
        return self.serialize(result)

    def modify_db_snapshot_attribute(self) -> TYPE_RESPONSE:
        params = self.parameters
        db_snapshot_identifier = params["DBSnapshotIdentifier"]
        db_snapshot_attributes_result = self.backend.modify_db_snapshot_attribute(
            db_snapshot_identifier=db_snapshot_identifier,
            attribute_name=params["AttributeName"],
            values_to_add=params.get("ValuesToAdd"),
            values_to_remove=params.get("ValuesToRemove"),
        )
        result = {
            "DBSnapshotAttributesResult": {
                "DBSnapshotIdentifier": db_snapshot_identifier,
                "DBSnapshotAttributes": [
                    {"AttributeName": name, "AttributeValues": values}
                    for name, values in db_snapshot_attributes_result.items()
                ],
            }
        }
        return self.serialize(result)

    def describe_db_cluster_snapshot_attributes(self) -> TYPE_RESPONSE:
        params = self.parameters
        db_cluster_snapshot_identifier = params["DBClusterSnapshotIdentifier"]
        db_cluster_snapshot_attributes_result = (
            self.backend.describe_db_cluster_snapshot_attributes(
                db_cluster_snapshot_identifier=db_cluster_snapshot_identifier,
            )
        )
        result = {
            "DBClusterSnapshotAttributesResult": {
                "DBClusterSnapshotIdentifier": db_cluster_snapshot_identifier,
                "DBClusterSnapshotAttributes": [
                    {"AttributeName": name, "AttributeValues": values}
                    for name, values in db_cluster_snapshot_attributes_result.items()
                ],
            }
        }
        return self.serialize(result)

    def modify_db_cluster_snapshot_attribute(self) -> TYPE_RESPONSE:
        params = self.parameters
        db_cluster_snapshot_identifier = params["DBClusterSnapshotIdentifier"]
        db_cluster_snapshot_attributes_result = (
            self.backend.modify_db_cluster_snapshot_attribute(
                db_cluster_snapshot_identifier=db_cluster_snapshot_identifier,
                attribute_name=params["AttributeName"],
                values_to_add=params.get("ValuesToAdd"),
                values_to_remove=params.get("ValuesToRemove"),
            )
        )
        result = {
            "DBClusterSnapshotAttributesResult": {
                "DBClusterSnapshotIdentifier": db_cluster_snapshot_identifier,
                "DBClusterSnapshotAttributes": [
                    {"AttributeName": name, "AttributeValues": values}
                    for name, values in db_cluster_snapshot_attributes_result.items()
                ],
            }
        }
        return self.serialize(result)

    def describe_db_proxies(self) -> TYPE_RESPONSE:
        params = self.parameters
        db_proxy_name = params.get("DBProxyName")
        # filters = params.get("Filters")
        marker = params.get("Marker")
        db_proxies = self.backend.describe_db_proxies(
            db_proxy_name=db_proxy_name,
            # filters=filters,
        )
        result = {
            "DBProxies": db_proxies,
            "Marker": marker,
        }
        return self.serialize(result)

    def create_db_proxy(self) -> TYPE_RESPONSE:
        params = self.parameters
        db_proxy_name = params["DBProxyName"]
        engine_family = params["EngineFamily"]
        auth = params["Auth"]
        role_arn = params["RoleArn"]
        vpc_subnet_ids = params["VpcSubnetIds"]
        vpc_security_group_ids = params.get("VpcSecurityGroupIds")
        require_tls = params.get("RequireTLS")
        idle_client_timeout = params.get("IdleClientTimeout")
        debug_logging = params.get("DebugLogging")
        tags = self.parameters.get("Tags", [])
        db_proxy = self.backend.create_db_proxy(
            db_proxy_name=db_proxy_name,
            engine_family=engine_family,
            auth=auth,
            role_arn=role_arn,
            vpc_subnet_ids=vpc_subnet_ids,
            vpc_security_group_ids=vpc_security_group_ids,
            require_tls=require_tls,
            idle_client_timeout=idle_client_timeout,
            debug_logging=debug_logging,
            tags=tags,
        )
        result = {"DBProxy": db_proxy}
        return self.serialize(result)

    def register_db_proxy_targets(self) -> TYPE_RESPONSE:
        db_proxy_name = self.parameters.get("DBProxyName")
        target_group_name = self.parameters.get("TargetGroupName")
        db_cluster_identifiers = self.parameters.get("DBClusterIdentifiers", [])
        db_instance_identifiers = self.parameters.get("DBInstanceIdentifiers", [])
        targets = self.backend.register_db_proxy_targets(
            db_proxy_name=db_proxy_name,
            target_group_name=target_group_name,
            db_cluster_identifiers=db_cluster_identifiers,
            db_instance_identifiers=db_instance_identifiers,
        )
        result = {"DBProxyTargets": targets}
        return self.serialize(result)

    def deregister_db_proxy_targets(self) -> TYPE_RESPONSE:
        db_proxy_name = self.parameters.get("DBProxyName")
        target_group_name = self.parameters.get("TargetGroupName")
        db_cluster_identifiers = self.parameters.get("DBClusterIdentifiers", [])
        db_instance_identifiers = self.parameters.get("DBInstanceIdentifiers", [])
        self.backend.deregister_db_proxy_targets(
            db_proxy_name=db_proxy_name,
            target_group_name=target_group_name,
            db_cluster_identifiers=db_cluster_identifiers,
            db_instance_identifiers=db_instance_identifiers,
        )
        return self.serialize({})

    def describe_db_proxy_targets(self) -> TYPE_RESPONSE:
        proxy_name = self.parameters.get("DBProxyName")
        targets = self.backend.describe_db_proxy_targets(proxy_name=proxy_name)
        result = {"Targets": targets}
        return self.serialize(result)

    def delete_db_proxy(self) -> TYPE_RESPONSE:
        proxy_name = self.parameters.get("DBProxyName")
        proxy = self.backend.delete_db_proxy(proxy_name=proxy_name)
        result = {"DBProxy": proxy}
        return self.serialize(result)

    def describe_db_proxy_target_groups(self) -> TYPE_RESPONSE:
        proxy_name = self.parameters.get("DBProxyName")
        groups = self.backend.describe_db_proxy_target_groups(proxy_name=proxy_name)
        result = {"TargetGroups": groups}
        return self.serialize(result)

    def modify_db_proxy_target_group(self) -> TYPE_RESPONSE:
        proxy_name = self.parameters.get("DBProxyName")
        config = self.parameters.get("ConnectionPoolConfig", {})
        group = self.backend.modify_db_proxy_target_group(
            proxy_name=proxy_name, config=config
        )
        result = {"DBProxyTargetGroup": group}
        return self.serialize(result)

    def describe_db_instance_automated_backups(self) -> TYPE_RESPONSE:
        automated_backups = self.backend.describe_db_instance_automated_backups(
            **self.parameters
        )
        result = {"DBInstanceAutomatedBackups": automated_backups}
        return self.serialize(result)

    def describe_events(self) -> TYPE_RESPONSE:
        events = self.backend.describe_events(**self.parameters)
        result = {"Events": events}
        return self.serialize(result)

    def describe_db_log_files(self) -> TYPE_RESPONSE:
        log_files = self.backend.describe_db_log_files(**self.parameters)
        result = {"DescribeDBLogFiles": log_files}
        return self.serialize(result)

    def _paginate(self, resources: List[Any]) -> Tuple[List[Any], Optional[str]]:
        from moto.rds.exceptions import InvalidParameterValue

        marker = self.parameters.get("Marker")
        # Default was originally set to 50 instead of 100 for ease of testing.  Should fix.
        page_size = self.parameters.get("MaxRecords", 50)
        if page_size < 20 or page_size > 100:
            msg = (
                f"Invalid value {page_size} for MaxRecords. Must be between 20 and 100"
            )
            raise InvalidParameterValue(msg)
        all_resources = list(resources)
        all_ids = [resource.resource_id for resource in all_resources]
        if marker:
            start = all_ids.index(marker) + 1
        else:
            start = 0
        paginated_resources = all_resources[start : start + page_size]
        next_marker = None
        if len(all_resources) > start + page_size:
            next_marker = paginated_resources[-1].resource_id
        return paginated_resources, next_marker



import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws

ONE_MB = 2**20


@mock_aws
def test_record_data_exceeds_1mb():
    client = boto3.client("kinesis", region_name="us-east-1")
    client.create_stream(StreamName="my_stream", ShardCount=1)
    with pytest.raises(ClientError) as exc:
        client.put_records(
            Records=[{"Data": b"a" * (ONE_MB + 1), "PartitionKey": "key"}],
            StreamName="my_stream",
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        err["Message"]
        == "1 validation error detected: Value at 'records.1.member.data' failed to satisfy constraint: Member must have length less than or equal to 1048576"
    )


@mock_aws
def test_record_data_and_partition_key_exceeds_1mb():
    client = boto3.client("kinesis", region_name="us-east-1")
    client.create_stream(StreamName="my_stream", ShardCount=1)

    key = "key"
    key_size = len(key)
    data_size = ONE_MB - key_size + 1
    with pytest.raises(ClientError) as exc:
        client.put_records(
            Records=[{"Data": b"a" * data_size, "PartitionKey": key}],
            StreamName="my_stream",
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        err["Message"]
        == "1 validation error detected: Value at 'records.1.member.data' failed to satisfy constraint: Member must have length less than or equal to 1048576"
    )


@mock_aws
def test_record_data_and_partition_key_exactly_1mb():
    client = boto3.client("kinesis", region_name="us-east-1")
    client.create_stream(StreamName="my_stream", ShardCount=1)

    key = "key"
    key_size = len(key)
    data_size = ONE_MB - key_size

    client.put_records(
        Records=[{"Data": b"a" * data_size, "PartitionKey": key}],
        StreamName="my_stream",
    )


@mock_aws
def test_record_data_and_partition_key_smaller_than_1mb():
    client = boto3.client("kinesis", region_name="us-east-1")
    client.create_stream(StreamName="my_stream", ShardCount=1)

    key = "key"
    key_size = len(key)
    data_size = ONE_MB - key_size - 1
    client.put_records(
        Records=[{"Data": b"a" * data_size, "PartitionKey": key}],
        StreamName="my_stream",
    )


@mock_aws
def test_total_record_data_exceeds_5mb():
    client = boto3.client("kinesis", region_name="us-east-1")
    client.create_stream(StreamName="my_stream", ShardCount=1)
    with pytest.raises(ClientError) as exc:
        client.put_records(
            Records=[{"Data": b"a" * 2**20, "PartitionKey": "key"}] * 5,
            StreamName="my_stream",
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidArgumentException"
    assert err["Message"] == "Records size exceeds 5 MB limit"


@mock_aws
def test_total_record_data_exact_5mb():
    client = boto3.client("kinesis", region_name="us-east-1")
    client.create_stream(StreamName="my_stream", ShardCount=1)
    key = "key"
    key_size = len(key)
    data_size = ONE_MB - key_size
    client.put_records(
        Records=[{"Data": b"a" * (data_size), "PartitionKey": key}] * 5,
        StreamName="my_stream",
    )


@mock_aws
def test_too_many_records():
    client = boto3.client("kinesis", region_name="us-east-1")
    client.create_stream(StreamName="my_stream", ShardCount=1)

    with pytest.raises(ClientError) as exc:
        client.put_records(
            Records=[{"Data": b"a", "PartitionKey": "key"}] * 501,
            StreamName="my_stream",
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        err["Message"]
        == "1 validation error detected: Value at 'records' failed to satisfy constraint: Member must have length less than or equal to 500"
    )
